# -*- coding: utf-8 -*-
# Editobj3
# Copyright (C) 2007-2014 Jean-Baptiste LAMY

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from cgi import escape as escape

def html_escape(s): return escape(s).replace("\n", "<br/>")

def html_escape_arg(s): return escape(s).replace('"', "&quot;")

def html_js_escape(s): return escape(s).replace('"', '\\"').replace("\n", "\\n")

def js_escape(s): return s.replace('"', '\\"').replace("\n", "\\n")

NEED_UPDATES = set()

class User(object):
  def __init__(self, name, password = ""):
    self.name        = name
    self.password    = password
    self._obj2id     = {}
    self._id2obj     = {}
    self.next_id     = 1
    self.ws_handlers = []
    self.urls        = {}
    USERS[name] = self

  def __repr__(self): return "<%s %s>" % (self.__class__.__name__, self.name)
    
  def get_user(self): return self
  
  def obj2id(self, obj):
    if not obj in self._obj2id:
      self._obj2id[obj              ] = str(self.next_id)
      self._id2obj[str(self.next_id)] = obj
      self.next_id += 1
    return self._obj2id[obj]
    
  def id2obj(self, id):
    return self._id2obj[id]
    
  def register_url(self, path, content):
    self.urls[path] = content

  def register_url_for_local_file(self, path, filename):
    self.urls[path] = URLForLocalFile(filename)
    

class URLForLocalFile(object):
  def __init__(self, filename):
    self.filename = filename

USERS = {}
GUEST = User("guest")



# -*- coding: utf-8 -*-
# Editobj3
# Copyright (C) 2007-2014 Jean-Baptiste LAMY

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
editobj3.observe: Observation framework
---------------------------------------

The observe module can register listeners for any Python object. When the object is modified,
the listeners will be (asynchronously) called. The following objects can be observed: Python instance,
lists and dictionaries. A listener is a function of the form::

  def my_listener(obj, type, new, old):
    ...

where obj is the modified object, type is the type of the modification, and old and new
are the old and new values. Type can be:
  
  - object (the Python root class): one or more attributes have changed on obj. old and
    new are the old and new attribute dictionary of obj (this dictionary includes attributes
    in obj.__dict__, but also Python's property and C-level getset).
    If you want to know which attribute has changed, use dictdiff on new and old (see the
    diffdict function docstring).
  
  - list : one or more addition / deletion have occured on obj (which is a list).
    new and old are the new and old list content of obj.

  - dict : one or more assignment / deletion have occured on obj (which is a dictionary).
    new and old are the new and old dictionary values.

  - "__class__" : the class of obj has changed. new and old are the new and old classes.

The :func:`scan()` function checks all observed objects and calls the corresponding listeners.

Quick example:
  >>> from editobj3.observe import *
  >>> class C: pass
  ...
  >>> c = C()
  >>> def listener(obj, type, new, old):
  ...   if type is object:
  ...     for (attr, newvalue, oldvalue) in diffdict(new, old):
  ...       print("c.%s was %s, is now %s" % (attr, oldvalue, newvalue))
  >>> observe(c, listener)
  >>> c.x = 1
  >>> scan()
  c.x was None, is now 1
"""

__all__ = [
  "observe", "isobserved", "unobserve", "observe_tree", "unobserve_tree",
  "scan", "diffdict", "Listener", "DebugListener",
  ]

from weakref import ref, WeakSet, WeakValueDictionary, WeakKeyDictionary

import editobj3

_observed_seqs    = {}
_observed_objects = {}

def _can_be_getter(method):
  if type(method).__name__ != "method": return False
  try: return method.__code__.co_argcount == 1
  except: return False

_class_2_properties = {}
def _get_class_properties(klass):
  props = _class_2_properties.get(klass)
  if props is None:
    props = _class_2_properties[klass] = [attr for attr in dir(klass) if (not attr in IGNORED_ATTRS) and (type(getattr(klass, attr)).__name__ in PROPERTY_TYPE_NAMES)]
  return props
  
_type_2_properties = {}
def _get_type_properties(typ):
  props = _type_2_properties.get(typ)
  if props is None:
    props = _type_2_properties[typ] = [attr for attr in dir(typ) if (not attr in IGNORED_ATTRS) and attr.startswith("get") and _can_be_getter(getattr(typ, attr))]
  return props

PROPERTY_TYPE_NAMES = { "property", "getset_descriptor" }
IGNORED_ATTRS       = { "__weakref__", "__abstractmethods__" }


class Listener(object):
  def __call__(self, obj, type, new, old):
    if   type is object:
      for attr, new_value, old_value in diffdict(new, old):
        self.attr_changed(obj, attr, new_value, old_value)
    elif type is list:
      for i in old: self.list_item_removed(obj, i)
      for i in new: self.list_item_added  (obj, i)
    elif type is dict:
      for k, i in old.items(): self.dict_item_removed(obj, k, i)
      for k, i in new.items(): self.dict_item_added  (obj, k, i)
    elif type == "__class__":
      self.class_changed(obj, new, old)
  def class_changed(self, obj, new_class, old_class): pass
  def attr_changed(self, obj, attr, new_value, old_value): pass
  def list_item_added  (self, obj, item): pass
  def list_item_removed(self, obj, item): pass
  def dict_item_added  (self, obj, key, item): pass
  def dict_item_removed(self, obj, key, item): pass

class DebugListener(Listener):
  def class_changed(self, obj, new_class, old_class): print("Class of '%s' changed from %s to %s." % (obj, old_class, new_class))
  def attr_changed(self, obj, attr, new_value, old_value): print("'%s.%s' was %s, is now %s." % (obj, attr, old_value, new_value))
  def list_item_added  (self, obj, item): print("'%s' added to '%s'." % (item, obj))
  def list_item_removed(self, obj, item): print("'%s' removed from '%s'." % (item, obj))
  def dict_item_added  (self, obj, key, item): print("XXX dict changed")
  def dict_item_removed(self, obj, key, item): print("XXX dict changed")


def observe(o, listener):
  """Registers a listener for the given object. When o will be changed, the listener will
be called (asynchronously). See the module docstrings for more info about what arguments receives a listener.

:param o: the object to observe.
:param listener: the listener.
:type listener: callable"""
  if isinstance(o, editobj3.introsp.ObjectPack):
    for o2 in o.objects: observe(o2, listener)
    return
  
  i = id(o)
  observation = _observed_seqs.get(i)
  if observation: observation.listeners.append(listener)
  else:
    if   isinstance(o, list): _observed_seqs[i] = ListObservation(o, [listener])
    elif isinstance(o, set ) or isinstance(o, WeakSet): _observed_seqs[i] = SetObservation (o, [listener])
    elif isinstance(o, dict) or isinstance(o, WeakValueDictionary) or isinstance(o, WeakKeyDictionary): _observed_seqs[i] = DictObservation(o, [listener])
    
  if hasattr(o, "__dict__"):
    observation = _observed_objects.get(i)
    if observation: observation.listeners.append(listener)
    else:           _observed_objects[i] = ObjectObservation(o, [listener])
    
def isobserved(o, listener = None):
  """Returns true if the given object is observed by the given listener. If listener is None, returns the list
of active listeners for o.

:param o: the object.
:param listener: the listener, or None.
:returns: True, False or list."""
  if isinstance(o, editobj3.introsp.ObjectPack):
    for o2 in o.objects:
      if isobserved(o2, listener): return True
    return False
    
  i = id(o)
  observation = _observed_seqs.get(i) or (hasattr(o, "__dict__") and _observed_objects.get(i))
  
  if listener: return observation and (listener in observation.listeners)
  else:        return observation and observation.listeners
  
def unobserve(o, listener = None):
  """Unregisters the given listener for o. If listener is not listening o,
nothing is done. If listener is None, unregisters *all* listeners on o.

:param o: the object.
:param listener: the listener, or None."""
  if isinstance(o, editobj3.introsp.ObjectPack):
    for o2 in o.objects: unobserve(o2, listener)
    return
    
  i = id(o)
  if listener:
    observation = _observed_seqs.get(i)
    if observation:
      try: observation.listeners.remove(listener)
      except ValueError: pass
      if not observation.listeners: del _observed_seqs[i]
      
    if hasattr(o, "__dict__"):
      observation = _observed_objects.get(i)
      if observation:
        try: observation.listeners.remove(listener)
        except ValueError: pass
        if not observation.listeners: del _observed_objects[i]
        
  else:
    if i in _observed_seqs   : del _observed_seqs   [i]
    if i in _observed_objects: del _observed_objects[i]
    

class Observation(object):
  def __init__(self, o, listeners):
    self.object    = o
    self.listeners = listeners
    self.old       = self.current_value()

class ListObservation(Observation):
  type = list
  def current_value(self): return list(self.object)
  
class SetObservation(Observation):
  type = set
  def current_value(self): return set(self.object)
  
class DictObservation(Observation):
  type = dict
  def current_value(self): return dict(self.object)
  
class ObjectObservation(Observation):
  def __init__(self, o, listeners):
    self.listeners = listeners
    if isinstance(o, type): self.props = _get_type_properties (o)
    else:                   self.props = _get_class_properties(o.__class__)
    self.old       = self.current_value(o)
    self.old_class = o.__class__
    try:    self.object = weakref.ref(o)
    except: self.object = o
    
  def current_value(self, o):
    if self.props:
      #new = { prop : getattr(o, prop, None) for prop in self.props }
      new = {}
      for prop in self.props:
        v = getattr(o, prop, None)
        if (type(v).__name__ == "method"): v = v()
        new[prop] = v
      new.update(o.__dict__)
      return new
    else: return o.__dict__.copy()

    
_has_owlready2 = False
    
def scan():
  """Checks for changes in all listened objects, and calls the corresponding listeners if needed."""
  for i, observation in list(_observed_seqs.items()):
    new = observation.object
    if   isinstance(new, WeakSet            ): new = set (new)
    elif isinstance(new, WeakKeyDictionary  ): new = dict(new)
    elif isinstance(new, WeakValueDictionary): new = dict(new)
    if observation.old != new:
      for listener in observation.listeners[:]: listener(observation.object, observation.type, new, observation.old)
      observation.old = observation.current_value()
      
  for i, observation in list(_observed_objects.items()):
    if type(observation.object) is ref:
      o = observation.object()
      if o is None:
        del _observed_objects[i]
        continue
    else: o = observation.object
    
    if observation.props:
      new = observation.current_value(o)
      
      if observation.old != new:
        for listener in observation.listeners[:]: listener(o, object, new, observation.old)
        observation.old = new
    else:
      if observation.old != o.__dict__:
        for listener in observation.listeners[:]: listener(o, object, o.__dict__, observation.old)
        observation.old = observation.current_value(o)
        
    if not observation.old_class is o.__class__:
      for listener in observation.listeners[:]: listener(o, "__class__", o.__class__, observation.old_class)
      observation.old_class = o.__class__
      
  if _has_owlready2:
    import owlready2.observe
    owlready2.observe.scan_collapsed_changes()
    

def diffdict(new, old, inexistent_value = None):
  """Returns the differences between two dictionaries.
In case of addition or deletion, old or new values are set to inexistent_value.

:param new: first dictionary to compare.
:param old: second dictionary to compare.
:param inexistent_value: value used as old or new value for deletion or addition, defaults to None.
:returns: a list of the differences between the two dictionaries, like: [(key, new_value, old_value),...]
"""
  changes = []
  
  for key, val in old.items():
    new_val = new.get(key, Ellipsis)
    if   new_val is Ellipsis: changes.append((key, inexistent_value, val))
    elif new_val != val:      changes.append((key, new_val, val))
    
  for key, val in new.items():
    old_val = old.get(key, Ellipsis)
    if old_val is Ellipsis: changes.append((key, val, inexistent_value))
    
  return changes


def find_all_children(o):
  if   isinstance(o, list): l = o
  elif isinstance(o, set) or isinstance(o, tuple) or isinstance(o, frozenset): l = list(o)
  elif isinstance(o, dict): l = list(o.keys()) + list(o.values())
  else:                     l = []
  if   hasattr(o, "__dict__"):
    if l is o: l = list(l)
    
    if isinstance(o, type): l += _get_type_properties (o)
    else:                   l += _get_class_properties(o.__class__)
    l += list(o.__dict__.values())
  return l
  
def observe_tree(o, listener, find_children = find_all_children):
  """Observes o with listener, as well as any item recursively in o (if o is a list, a dict,
or have attributes). Items added to or removed from o
or one of its items are automatically observed or unobserved.
Although called "observe_tree", it works with any nested structure of lists and dicts,
including cyclic ones.

You must use :func:`unobserve_tree` to remove the listener.

:param o: the object to observe.
:param listener: the listener.
:type listener: callable
:param find_children: optional function to call for listing the children of o.
"""
  _observe_tree(o, _TreeListener(o, listener, find_children))
  
def _observe_tree(o, listener):
  if not isobserved(o, listener): # Avoid troubles with cyclic list / dict
    observe(o, listener)
    
    children = listener.find_children(o)
    for child in children:
      _observe_tree(child, listener)
      
def unobserve_tree(o, listener, find_children = find_all_children, already = None):
  """Unregisters the given tree listener for o.

:param o: the object to observe.
:param listener: the listener, or None for unregistering *all* listeners.
:param find_children: optional function to call for listing the children of o.
"""
  if already is None: already = set()
  
  if not id(o) in already:
    already.add(id(o))
    unobserve(o, listener)
    
  children = find_children(o)
  for child in children:
    if not id(child) in already:
      unobserve_tree(child, listener, find_children, already)
      
class _TreeListener:
  def __init__(self, o, listener, find_children = find_all_children):
    self.object        = o
    self.listener      = listener
    self.find_children = find_children
    
  def __eq__(self, other): return other == self.listener
  
  def __call__(self, o, type, new, old):
    if   type is list:
      for item in old:
        if not item in new: unobserve_tree(item, self)
        
      for item in new:
        if not item in old: _observe_tree (item, self)
        
    elif (type is dict) or (type is object):
      _new = new.values()
      _old = old.values()
      for item in _old:
        if not item in _new: unobserve_tree(item, self)

      for item in _new:
        if not item in _old: _observe_tree (item, self)
        
    self.listener(o, type, new, old)




# -*- coding: utf-8 -*-
# Editobj3
# Copyright (C) 2007-2014 Jean-Baptiste LAMY

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, os, os.path, urllib.parse, urllib.request, json, base64
from io import StringIO
#try:    import ssl
#except: ssl = None

import asyncio, aiohttp, aiohttp.web, aiohttp.http_websocket

import editobj3, editobj3.field_html
#from editobj3.http_ws_server  import *
from editobj3.html_utils      import *
from editobj3.editor_html     import *
import editobj3.observe as observe
import editobj3.introsp as introsp

editobj3.eval = lambda x: x # Security !

JS_DIR = os.path.join(os.path.dirname(editobj3.__file__), "js")
if not os.path.exists(JS_DIR):
  JS_DIR = os.path.join("/usr", "share", "editobj3", "js")
  if not os.path.exists(JS_DIR):
    JS_DIR = os.path.join("/usr", "local", "share", "editobj3", "js")
    if not os.path.exists(JS_DIR):
      JS_DIR = os.path.join("/usr", "share", "python-editobj3", "js")
      
APPLICATION = None


class EditobjServer(aiohttp.web.Application):
  def __init__(self, on_startup = None, *args, **kargs):
    aiohttp.web.Application.__init__(self, *args, **kargs)
    self.commit_task = None
    
    if on_startup:
      def on_startup2(app): self.loop.call_later(0.05, on_startup)
      self.on_startup.append(on_startup2)
    self.router.add_static("/editobj/js/", JS_DIR)
    self.router.add_get   ("/editobj/icons/{icon}", self.on_icons)
    self.router.add_get   ("/editobj/editor.html", self.on_editor)
    self.router.add_get   ("/editobj/editor.ws", self.on_open_websocket)
    self.router.add_get   ("/editobj/{obj}", self.on_other)
    self.router.add_get   ("/editobj/auth/", self.on_auth)
    
  EXT_2_CONTENT_TYPE = {
    "html" : "text/html",
    "css"  : "text/css",
    "js"   : "text/javascript",
    "xml"  : "text/xml",
    "txt"  : "text/plain",
    "py"   : "text/python",
    "png"  : "image/png",
    "gif"  : "image/gif",
    "jpeg" : "image/jpeg",
    "jpg"  : "image/jpeg",
    "svg"  : "image/svg+xml",
  }
  def guess_content_type(self, filename):
    return self.EXT_2_CONTENT_TYPE.get(filename.rsplit(".")[-1], "text/plain")
    
  def get_current_user(self, request):
    auth_header = request.headers.get("Authorization", "")
    if auth_header.strip() == "": return GUEST
    auth_header = auth_header[6:].strip() # 6 == len("Basic ")
    user_name, password = base64.b64decode(auth_header).decode("latin").split(":")
    user = USERS.get(user_name)
    if user and (user.password == password): return user
    return GUEST
    
  async def on_auth(self, request):
    #elif (user is GUEST) and path.startswith("/auth/"):
    return aiohttp.web.Response(status = 401, headers = { "WWW-Authenticate" : 'Basic realm="editobj"' })
  
  async def on_icons(self, request):
    format, icon = url_2_icon["/editobj/icons/%s" % request.match_info.get("icon")]
    #format = self.EXT_2_CONTENT_TYPE[format]
    if isinstance(icon, str):
      #return aiohttp.web.Response(content_type = format, body = open(icon, "rb").read())
      return aiohttp.web.FileResponse(icon)
    else:
      return aiohttp.web.Response(content_type = format, body = icon)
    
  async def on_editor(self, request):
    user   = self.get_current_user(request)
    editor = user.id2obj(request.query["id"])
    return aiohttp.web.Response(content_type = "text/html", text = html_page_for_editor(editor, self.server_host, self.server_port))
  
  async def on_other(self, request):
    user = self.get_current_user(request)
    
    if (request.path in user.urls) or (request.path in GUEST.urls):
      obj = user.urls.get(request.path) or GUEST.urls.get(request.path)
      if   isinstance(obj, BaseDialog):
        return aiohttp.web.Response(content_type = "text/html", text = html_page_for_editor(obj, self.server_host, self.server_port))
      elif hasattr(obj, "get_html"):
        return aiohttp.web.Response(content_type = "text/html", text = obj.get_html())
      elif isinstance(obj, URLForLocalFile):
        content_type = self.guess_content_type(obj.filename)
        if content_type.startswith("text"):
          return aiohttp.web.Response(content_type = "text/html", text = open(obj.filename).read())
        else:
          return aiohttp.web.Response(content_type = "text/html", text = open(obj.filename, "rb").read())

    if user is GUEST:
      return aiohttp.web.Response(status = 401, headers = { "WWW-Authenticate" : 'Basic realm="editobj"' })
      
    raise ValueError(user, request.path)

  async def on_open_websocket(self, request):
    user      = self.get_current_user(request)
    websocket = EditobjWebSocket(self, user, request)
    return await websocket.run()
  
  def run(self, host = "127.0.0.1", port = 8080, *args, **kargs):
    global APPLICATION
    
    self.server_host = host
    self.server_port = port
    observe.scan = self.scan
    APPLICATION  = self
    
    aiohttp.web.run_app(self, host = host, port = port, *args, **kargs)
    
  def scan(self):
    _observe_scan()
    self.schedule_commit()
    
  def schedule_commit(self):
    if self.commit_task: self.commit_task.cancel()
    self.commit_task = self.loop.create_task(self.commit())
    #self.loop.call_later(0.1, self.commit_task)
    
  async def commit(self):
    updated_editors = set()
    for user in USERS.values():
      for ws_handler in user.ws_handlers:
        js = ""
        if ws_handler.editor.get_need_update():
          js += ws_handler.editor.get_update_js()
          updated_editors.add(ws_handler.editor)
        if ws_handler.dialog and ws_handler.dialog.get_need_update():
          js += ws_handler.dialog.get_update_js()
          updated_editors.add(ws_handler.dialog)
        if js:
          await ws_handler.send_ws_data(js)

    for editor in updated_editors: editor.clear_update()

_observe_scan = observe.scan


class EditobjWebSocket(object):
  def __init__(self, application, user, request):
    self.application = application
    self.user        = user
    self.request     = request
    self.response    = aiohttp.web.WebSocketResponse()
    self.editor      = user.id2obj(request.query["id"])
    self.dialog      = None
    
    self.user.ws_handlers.append(self)
    
  async def run(self):
    await self.response.prepare(self.request)
    
    async for msg in self.response:
      if msg.type == aiohttp.WSMsgType.TEXT:
        await self.on_ws_data_received(msg.data)
        
      elif msg.type == aiohttp.WSMsgType.ERROR:
        print('ws connection closed with exception %s' % self.response.exception())
        self.on_closed()
        break
        
    self.on_closed()
    return self.response
  
  async def on_ws_data_received(self, data):
    data = json.loads(data)
    
    if   data[0] == "setattr":
      field = self.user.id2obj(data[1])
      field.set_value(data[2])
      
    elif data[0] == "click":
      field = self.user.id2obj(data[1])
      field.on_click()
      self.application.scan()
      
    elif data[0] == "closed":
      dialog = self.user.id2obj(data[1])
      dialog.on_closed()
      self.application.scan()
      
    elif data[0] == "getnode":
      if data[2] == "#":
        hierarchy_pane = self.user.id2obj(data[1])
        j = "[%s]" % hierarchy_pane.root_node.get_json()
      else:
        parent = self.user.id2obj(data[2])
        j = parent.get_children_node_json()
      await self.send_ws_data("""data_received(%s, %s);""" % (j, data[3]))
      
    elif data[0] == "editnodes":
      editor =  self.user.id2obj(data[1])
      nodes  = [self.user.id2obj(id) for id in data[2:]]
      editor.on_selection_changed(nodes)
      await self.application.commit()
      
    elif data[0] == "nodeaction":
      node = self.user.id2obj(data[1])
      node.do_action_by_label(data[2])
      
    elif data[0] == "tbclick":
      editor = self.user.id2obj(data[1])
      editor.on_click(data[2])
      
    elif data[0] == "tab":
      editor = self.user.id2obj(data[1])
      editor.on_tab(data[2], data[3])
      await self.application.commit()
      
  async def send_ws_data(self, data):
    await self.response.send_str(data)
    
  def on_closed(self):
    self.user.ws_handlers.remove(self)
    




def html_page_for_editor(editor, host, port):
    if   isinstance(editor, HTMLEditorDialog):
      if editor.menubar: resizer = """
function resize_full_height() { BYID("%s").style.height = (window.innerHeight - 6 - $("#tb%s").height()) + "px"; }
""" % (editor.editor_pane._id, editor._id)
      else: resizer = """
function resize_full_height() { BYID("%s").style.height = (window.innerHeight - 1) + "px"; }
""" % (editor.editor_pane._id)
      
    elif isinstance(editor, HTMLEditorTabbedDialog):
      if editor.menubar: resizer = """function resize_full_height() { var h = (window.innerHeight - 10 - $("#tabs%s").height() - $("#tb%s").height()) + "px";\n""" % (editor._id, editor._id)
      else:              resizer = """function resize_full_height() { var h = (window.innerHeight - 3 - $("#tabs%s").height()) + "px";\n""" % editor._id
      for editor_pane in editor.tabs:
        resizer += """  if(BYID("%s")) BYID("%s").style.height = h;\n""" % (editor_pane._id, editor_pane._id)
      resizer += "}\n"
    else: resizer = """
function resize_full_height() { BYID("%s").style.height = (window.innerHeight - 1) + "px"; }
""" % editor._id

    html = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
  <head>
    <title>%s</title>
    <link rel="stylesheet" href="/editobj/js/w2ui-1.3.2.css" />
    <script type="text/javascript" src="/editobj/js/jquery-1.11.0.min.js"></script>
    <script type="text/javascript" src="/editobj/js/w2ui-1.3.2.js"></script>
    <style type='text/css'>
.w2ui-icon { width: %spx; height: %spx; }
.w2ui-sidebar .w2ui-sidebar-div .w2ui-node .w2ui-node-image { width: %spx; height: %spx; }
.w2ui-sidebar .w2ui-sidebar-div .w2ui-node-caption { padding-top: 7px; }
.w2ui-sidebar .w2ui-sidebar-div td.w2ui-node-data { padding: 0px; }
.w2ui-sidebar .w2ui-sidebar-div td.w2ui-node-data .w2ui-node-image.w2ui-icon { margin-top: 0px; margin-bottom: 0px; }
.icon-add2 { background: url(/editobj/icons/1) no-repeat center; background-size: contain; }
.icon-remove { background: url(/editobj/icons/2) no-repeat center; background-size: contain; }
.icon-up { background: url(/editobj/icons/3) no-repeat center; background-size: contain; }
.icon-down { background: url(/editobj/icons/4) no-repeat center; background-size: contain; }
.editobj-button {
  background-color: #FAFAFA;
  text-align: center;
  border: 1px solid #AAAAAA;
  border-radius: 3px;
  cursor: default;
 }
.editobj-button:hover {
  background-color: #FFFFFF;
 }
.editobj-button:active {
  background-color: #EEEEEE;
 }
input[type=checkbox] { -ms-transform: scale(1.5); -moz-transform: scale(1.5); -webkit-transform: scale(1.5); -o-transform: scale(1.5); padding: 10px; }
    </style>
    <script type="text/javascript">
function BYID(id) { return document.getElementById(id) };
ws = new WebSocket("ws://%s:%s/editobj/editor.ws?id=%s");
ws.onmessage = function(event) {
  eval(event.data);
};
queue = new Array();
next_queue_id = 1;
function ws_send_and_wait(data, callback) {
  if (ws.readyState == 0) {
    setTimeout(function() { ws_send_and_wait(data, callback); }, 100);
  }
  else {
    var queue_id = next_queue_id;
    next_queue_id = next_queue_id + 1;
    data.push(queue_id);
    ws.send(JSON.stringify(data));
    queue[queue_id] = callback;
  }
}
function data_received(json, queue_id) {
  queue[queue_id](json);
  delete queue[queue_id];
}
function create_css_class(classname, content) {
  var style = document.createElement("style");
  style.type = "text/css";
  style.innerHTML = "." + classname + " {" + content + "}";
  document.getElementsByTagName("head")[0].appendChild(style);
}
%s
$(window).resize(resize_full_height);

%s
    </script>
  </head>
  <body style="margin: 0px; background: #EEEEEE;">
%s

<script>
$(function () {
%s
resize_full_height();
});
</script>
  </body>
</html>
""" % (html_escape(introsp.description_for_object(editor.o).label_for(editor.o)), SMALL_ICON_SIZE, SMALL_ICON_SIZE, SMALL_ICON_SIZE, SMALL_ICON_SIZE, host, port, editor._id, resizer, editobj3.field_html.FIELD_JS + editobj3.editor_html.EDITOR_JS, editor.get_html(), editor.get_ready_js())
    return html




# -*- coding: utf-8 -*-
# Editobj3
# Copyright (C) 2007-2014 Jean-Baptiste LAMY

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import string
import editobj3.introsp as introsp
from editobj3.editor     import *
from editobj3.html_utils import *

EDITOR_JS = ""

class Menu(object):
  def __init__(self, master, label, command, arg, image, type, radio_group):
    self._id = master.get_user().obj2id(self)
    self.submenus    = []
    self.label       = editobj3.TRANSLATOR(label)
    self.command     = command
    self.arg         = arg
    self.image       = image
    self.type        = type
    self.radio_group = radio_group
  
  def __len__(self): return len(self.submenus)
  
  def insert(self, pos, menu): self.submenus.insert(pos, menu)
  
  def on_click(self):
    if self.command:
      if self.arg: self.command(self.arg)
      else:        self.command()
      
  def get_json(self):
    l = ['id:"%s"' % self._id, 'caption:"%s"' % js_escape(self.label)]
    if self.submenus:
      l.append('type:"menu"')
      l.append('items:[%s]' % ",".join(submenu.get_json() for submenu in self.submenus))
    else: l.append('type:"%s"' % self.type)
    if self.image: l.append('icon:"icon%s"' % (load_small_icon(self.image).rsplit("/", 1)[-1].split(".", 1)[0]))
    if self.type == "radio": l.append('group:"%s"' % self.radio_group)
    return "{%s}" % ",".join(l)
    
    
class BaseDialog(object):
  def build_default_menubar(self):
    self.next_radio_group = 1
    self.current_radio_group = None
    
    self.file_menu = self.add_to_menu(self.menubar  , 1, u"File")
    #self.close_menu= self.add_to_menu(self.file_menu, 0, u"Close", self.on_ok)
    self.edit_menu = self.add_to_menu(self.menubar  , 1, u"Edit") # , self.on_edit_menu) # Not yet supported by W2UI for submenu
    self.undo_menu = self.add_to_menu(self.edit_menu, 0, u"Undo" , self.on_undo, accel = u"C-Z")
    self.redo_menu = self.add_to_menu(self.edit_menu, 0, u"Redo" , self.on_redo, accel = u"C-Y")
    if self.on_validate:
      self.validate_menu = self.add_to_menu(self.menubar, 0, u"Validate", self.on_click)
      
  def add_to_menu(self, menu, has_submenu, label, command = None, arg = None, accel = "", accel_enabled = 1, image = None, type = "button", pos = -1):
    if type == "radio":
      if self.current_radio_group is None:
        self.current_radio_group = self.next_radio_group
        self.next_radio_group += 1
    else: self.current_radio_group = None
    if pos == -1: pos = len(menu)
    new_menu = Menu(self, label, command, arg, image, type, self.current_radio_group)
    menu.insert(pos, new_menu)
    return new_menu
    
  def add_separator_to_menu(self, menu, pos = -1):
    self.current_radio_group = None
    if pos == -1: pos = len(menu)
    new_menu = Menu(self, "", None, None, None, "break", 0)
    menu.insert(pos, new_menu)
    return new_menu
    
  def get_menubar_json(self):
    return """
onClick: function (event) {
if (event.subItem != null)
     ws.send(JSON.stringify(['click', event.subItem.id]));
else ws.send(JSON.stringify(['click', event.target]));
}, items: [%s]""" % ",".join(menu.get_json() for menu in self.menubar)
  
  def set_menu_enable(self, menu, enable):
    self.update_js += """tb%s.%sable("%s")\n""" % (self._id, "en" if enable else "dis", menu._id)
    
  def set_menu_label (self, menu, label):
    self.update_js += """tb%s.set("%s", {caption: "%s"})\n""" % (self._id, menu._id, label)
    
  # Not yet supported by W2UI for submenu
  #def on_edit_menu(self, *args):
  #  undo = self.undo_stack.can_undo()
  #  redo = self.undo_stack.can_redo()
  #  if undo: self.set_menu_enable(self.undo_menu, 1); self.set_menu_label(self.undo_menu, "%s %s" % (editobj3.TRANSLATOR(u"Undo"), undo.name))
  #  else:    self.set_menu_enable(self.undo_menu, 0); self.set_menu_label(self.undo_menu, editobj3.TRANSLATOR(u"Undo"))
  #  if redo: self.set_menu_enable(self.redo_menu, 1); self.set_menu_label(self.redo_menu, "%s %s" % (editobj3.TRANSLATOR(u"Redo"), redo.name))
  #  else:    self.set_menu_enable(self.redo_menu, 0); self.set_menu_label(self.redo_menu, editobj3.TRANSLATOR(u"Redo"))
    
  def on_undo(self, *args):
    if self.undo_stack.can_undo(): self.undo_stack.undo()
  def on_redo(self, *args):
    if self.undo_stack.can_redo(): self.undo_stack.redo()
    
  def on_ok(self, *args):
    if self.on_close: self.on_close()
    
  def main(self, host = "127.0.0.1", port = 8080, path = "/editobj/main.html"):
    import webbrowser, editobj3.html_server
    self.master.register_url(path, self)
    def on_startup(): webbrowser.open_new(url)
    server = editobj3.html_server.EditobjServer(on_startup = on_startup)
    url    = "http://%s:%s%s" % (host, port, path)
    print("* Editobj3 * Editor dialog ready at %s !" % url)
    server.run(host, port)
    
  default_width  = 1024
  default_height =  768
  def set_default_size(self, width, height):
    self.default_width  = width
    self.default_height = height
    
  def show(self):
    user = self.get_user()
    root_editor = self
    while not isinstance(root_editor.master, User): root_editor = root_editor.master
    for ws_handler in user.ws_handlers:
      if ws_handler.editor is root_editor:
        ws_handler.dialog = self
        self.ws_handler = ws_handler
        break
    else:
      if self.is_dialog: # Dialog => must have a root editor!
        raise ValueError("Cannot find root editor / WS handler for dialog!")
        
    import editobj3.html_server
    if editobj3.html_server.APPLICATION: editobj3.html_server.APPLICATION.schedule_commit()
    
  def on_click(self):
    self.validated = True
    self.on_validate(self.get_selected_object())
    
  def on_closed(self):
    if self.on_validate and not self.validated: self.on_validate(None)
    if not self.ws_handler is None:
      self.ws_handler.dialog = None
      self.ws_handler = None
      

class HTMLEditorDialog(EditorDialog, BaseDialog):
  def __init__(self, gui = None, master = None, direction = "h", on_validate = None, edit_child_in_self = 1, undo_stack = None, on_close = None, menubar = True):
    self.master      = master or GUEST # Default to guest user
    self.menubar     = []
    super().__init__(gui, master, direction, on_validate, edit_child_in_self, undo_stack, on_close, menubar)
    self._id = self.master.get_user().obj2id(self)
    self.is_dialog   = not isinstance(self.master, User)
    self.on_close    = on_close
    self.need_update = True
    self.validated   = False
    self.update_js   = ""
    
  def get_user(self): return self.master.get_user()
  
  def get_o(self): return self.editor_pane.o
  o = property(get_o)
  
  def get_html(self):
    self.need_update = False
    if self.is_dialog: return ""
    else:
      return """
<div id="%s">
%s
</div>
""" % (self._id, self.get_inner_html())
  
  def get_inner_html(self):
    if self.menubar:
      html = """<div id="tb%s" style="padding: 2px; border-bottom: 1px solid #aaaaaa;  box-shadow: 0px 0px 5px #999; position: relative; z-index: 999;">MENUBAR</div>\n""" % self._id
    else: html = ""
    return html + self.editor_pane.get_html()
    
  READY_JS_TEMPLATE = string.Template("""
dialog$id = w2popup.open({ title: "$title", showMax: true, showClose: true, modal: $modal, buttons: "$buttons", body: "$body",
  onOpen : function (event) { event.onComplete = function () {
    $ready_js
    BYID("$editor_pane_id").style.height = (window.innerHeight - $delta_y) + "px";
    layout$editor_pane_id.resize();
  }},
  onMax: function (event) { event.onComplete = function () { layout$editor_pane_id.resize(); }},
  onMin: function (event) { event.onComplete = function () { layout$editor_pane_id.resize(); }},
  onClose: function (event) { ws.send(JSON.stringify(['closed', '$id'])); },
  width: window.innerWidth * 0.8, height: window.innerHeight
});
""")
  def get_ready_js(self):
    if self.is_dialog:
      o = self.o
      title = introsp.description_for_object(o).label_for(o).replace("\n", " ").replace("\t", " ")
      if self.on_validate:
        buttons  = """<input type="button" value="%s" onclick="w2popup.close();"/>&nbsp;&nbsp;&nbsp;""" % editobj3.TRANSLATOR("Cancel")
        buttons += """<input type="button" value="%s" onclick="ws.send(JSON.stringify(['click', '%s'])); w2popup.close();"/>""" % (editobj3.TRANSLATOR("Validate"), self._id)
      else: buttons = ""
      js = self.READY_JS_TEMPLATE.substitute(id = self._id, editor_pane_id = self.editor_pane._id,
                                             title = html_js_escape(title), modal = "true" if self.on_validate else "false",
                                             buttons = js_escape(buttons), body = js_escape(self.get_inner_html()),
                                             ready_js = self.editor_pane.get_ready_js(),
                                             width = self.default_width, height = self.default_height,
                                             delta_y = 120 if self.menubar else 100)
    else: js = self.editor_pane.get_ready_js()
    if self.menubar: js += """
tb%s = $('#tb%s').w2toolbar({ name: 'tb%s', %s});
""" % (self._id, self._id, self._id, self.get_menubar_json())
    return js
    
  def get_need_update(self): return self.need_update or self.update_js or self.editor_pane.get_need_update()
  
  def get_update_js(self):
    if self.need_update:
      if self.is_dialog: return self.get_ready_js()
      else: return """window.open("/editobj/editor.html?id=%s", "_blank");""" % self._id
    else: return self.editor_pane.get_update_js() + self.update_js
      
  def clear_update(self):
    self.need_update = False
    self.update_js   = ""
    self.editor_pane.clear_update()
    
  
class HTMLEditorTabbedDialog(EditorTabbedDialog, BaseDialog):
  def __init__(self, gui = None, master = None, direction = "h", on_validate = None, edit_child_in_self = 1, undo_stack = None, on_close = None, menubar = True):
    self.master      = master or GUEST # Default to guest user
    self.menubar     = []
    super().__init__(gui, master, direction, on_validate, edit_child_in_self, undo_stack, on_close, menubar)
    self._id = self.master.get_user().obj2id(self)
    self.is_dialog   = not isinstance(self.master, User)
    self.on_close    = on_close
    self.need_update = True
    self.validated   = False
    self.update_js   = ""
    self.current_tab = 0
    self.tab_labels  = []
    self.tabs        = []
    
  def add_tab(self, name, label, o = None):
    # XXX does not support adding / removing tabs after showing the dialog box / page
    editor_pane = super().add_tab(name, label, o)
    self.tab_labels.append(label)
    self.tabs.append(editor_pane)
    return editor_pane
    
  def remove_tab(self, name):
    # XXX does not support adding / removing tabs after showing the dialog box / page
    self.tab_labels.remove(name)
    self.tabs.remove(self.editor_panes[names])
    super.remove_tab(name)
    
  def on_tab(self, tab, already_sent):
    old_tab = self.current_tab
    self.current_tab = int(tab)
    current = self.get_current_editor_pane()
    if already_sent is None:
      self.update_js += """$("#tab%s_%s").html("%s");""" % (self._id, self.current_tab, js_escape(current.get_html()))
      self.update_js += current.get_ready_js()
      self.update_js += """tabs_received_%s["%s"] = true;""" % (self._id, self.current_tab)
    self.update_js += """BYID("tab%s_%s").style.display = "none";"""  % (self._id, old_tab)
    self.update_js += """BYID("tab%s_%s").style.display = "block";""" % (self._id, self.current_tab)
    self.update_js += """resize_full_height();\n"""
    
  def get_current_editor_pane(self): return self.tabs[self.current_tab]
  
  def get_user(self): return self.master.get_user()
  
  def get_o(self): return self.get_current_editor_pane().o
  o = property(get_o)
  
  def get_html(self):
    self.need_update = False
    if self.is_dialog: return ""
    else:
      return """
<div id="%s">
%s
</div>
""" % (self._id, self.get_inner_html())
  
  def get_inner_html(self):
    if self.menubar:
      html = """<div id="tb%s" style="padding: 2px; box-shadow: 0px 0px 5px #999;">MENUBAR</div>\n""" % self._id
    else: html = ""
    html += """
<div id="tabs%s" style="width: 100%%;">TABS</div>
""" % (self._id)
    for i in range(len(self.editor_panes)):
      if i == self.current_tab:
        html += """<div id="tab%s_%s">%s</div>""" % (self._id, i, self.get_current_editor_pane().get_html())
      else:
        html += """<div id="tab%s_%s" style="display: none;">TAB</div>""" % (self._id, i)
    return html

  READY_JS_TEMPLATE = string.Template("""
dialog$id = w2popup.open({ title: "$title", showMax: true, showClose: true, modal: $modal, buttons: "$buttons", body: "$body",
  onOpen : function (event) { event.onComplete = function () {
    $ready_js
    BYID("$editor_pane_id").style.height = (window.innerHeight - $delta_y) + "px";
    layout$editor_pane_id.resize();
  }},
  onMax: function (event) { event.onComplete = function () { layout$editor_pane_id.resize(); }},
  onMin: function (event) { event.onComplete = function () { layout$editor_pane_id.resize(); }},
  onClose: function (event) { ws.send(JSON.stringify(['closed', '$id'])); },
  width: window.innerWidth * 0.8, height: window.innerHeight
});
""")
  def get_ready_js(self):
    if self.is_dialog:
      o = self.o
      title = introsp.description_for_object(o).label_for(o).replace("\n", " ").replace("\t", " ")
      if self.on_validate:
        buttons  = """<input type="button" value="%s" onclick="w2popup.close();"/>&nbsp;&nbsp;&nbsp;""" % editobj3.TRANSLATOR("Cancel")
        buttons += """<input type="button" value="%s" onclick="ws.send(JSON.stringify(['click', '%s'])); w2popup.close();"/>""" % (editobj3.TRANSLATOR("Validate"), self._id)
      else: buttons = ""
      js = self.READY_JS_TEMPLATE.substitute(id = self._id, editor_pane_id = self.get_current_editor_pane()._id,
                                             title = html_js_escape(title), modal = "true" if self.on_validate else "false",
                                             buttons = js_escape(buttons), body = js_escape(self.get_inner_html()),
                                             ready_js = self.get_current_editor_pane().get_ready_js(),
                                             width = self.default_width, height = self.default_height,
                                             delta_y = 120 if self.menubar else 100)
    else: js = self.get_current_editor_pane().get_ready_js()
    if self.menubar: js += """
tb%s = $('#tb%s').w2toolbar({ name: 'tb%s', %s});
""" % (self._id, self._id, self._id, self.get_menubar_json())
    js += """
tabs_received_%s = new Array();
tabs_received_%s["%s"] = true;
$("#tabs%s").w2tabs({name: "tabs%s", active: "%s", tabs: [%s], style:"background-color: transparent;",
  onClick: function (event) { ws.send(JSON.stringify(["tab", "%s", event.target, tabs_received_%s[event.target]])); }
});
""" % (self._id, self._id, self.current_tab, self._id, self._id, self.current_tab, ",".join('{id:"%s", caption:"%s"}' % (i, js_escape(self.tab_labels[i])) for i in range(len(self.editor_panes))), self._id, self._id)
    return js
    
  def get_need_update(self): return self.need_update or self.update_js or self.get_current_editor_pane().get_need_update()
  
  def get_update_js(self):
    if self.need_update:
      if self.is_dialog: return self.get_ready_js()
      else: return """window.open("/editobj/editor.html?id=%s", "_blank");""" % self._id
    else: return self.update_js + self.get_current_editor_pane().get_update_js()
    
  def clear_update(self):
    self.need_update = False
    self.update_js   = ""
    self.get_current_editor_pane().clear_update()
    
  
class HTMLEditorPane(EditorPane):
  def __init__(self, gui, master, edit_child_in_self = 1, undo_stack = None, direction = "h"):
    super().__init__(gui, master, edit_child_in_self, undo_stack)
    self._id = master.get_user().obj2id(self)
    
  def get_user(self): return self.master.get_user()
  
  def get_o(self): return self.attribute_pane.o
  o = property(get_o)
  
  def get_html(self):
    return """
<div id="%s" style="width: 100%%; height: 752px;"></div>
<span id="editor%s">
%s
<div style="padding-top: 1em; clear:left;">%s</div>
</span>
""" % (self._id, self._id, self.icon_pane.get_html(), self.attribute_pane.get_html())
    
  def get_inner_html(self):
    return """
<span id="editor%s">
%s
<div style="padding-top: 1em; clear:left;">%s</div>
</span>
""" % (self._id, self.icon_pane.get_html(), self.attribute_pane.get_html())

  READY_JS_TEMPLATE = string.Template("""
$layout = $$("#$id").w2layout(
{ name: "$layout", padding: 0, panels: [
{ type: "left", minSize: 0, style: "border: 1px solid #AAAAAA; border-top: none;", size: window.innerWidth * 0.4, resizable: true, toolbar: { $toolbars style: "border: 1px solid #aaaaaa; border-top:none;" } },
{ type: "main", minSize: 550, style: "background: #EEEEEE; padding: 10px;" }
]});
$layout.content("left", $tree);
$layout.content("main", BYID("editor$id"));
$tb = $layout.panels[0].toolbar;
$tb.tree = $tree;
""")
  
  def get_ready_js(self, self_id = ""):
    if not self_id: self_id = "layout%s" % self._id
    js = self.attribute_pane.get_ready_js() + self.hierarchy_pane.get_ready_js() + self.READY_JS_TEMPLATE.substitute(id = self._id, layout = "%s" % self_id, toolbars = self.childhood_pane.get_toolbar_items(), tree = "tree%s" % self.hierarchy_pane._id, tb = "tb%s" % self.childhood_pane._id)
    if not self.hierarchy_visible:
      js += """layout%s.hide("left", true);""" % self._id
    return js
    
  def get_need_update(self): return self.icon_pane.get_need_update() or self.attribute_pane.get_need_update() or self.hierarchy_pane.get_need_update() or self.childhood_pane.get_need_update()
  
  def get_update_js(self):
    return "\n".join([self.icon_pane     .get_update_js(),
                      self.attribute_pane.get_update_js(),
                      self.hierarchy_pane.get_update_js(),
                      self.childhood_pane.get_update_js(),
                      ]).strip()
    
  def clear_update(self):
    self.icon_pane     .clear_update()
    self.attribute_pane.clear_update()
    self.hierarchy_pane.clear_update()
    self.childhood_pane.clear_update()
    
  def _set_hierarchy_visible(self, visible):
    self.hierarchy_visible = visible
     

class HTMLAttributePane(AttributePane):
  def __init__(self, gui, master, edit_child = None, undo_stack = None):
    super().__init__(gui, master, edit_child, undo_stack)
    self.labels_fields_units = []
    self._id = master.get_user().obj2id(self)
    self.need_update  = False
  
  def get_need_update(self): return self.need_update
    
  def edit(self, o, ignored_attrs = None):
    if self.o is o: return
    AttributePane.edit(self, o, ignored_attrs)
    
  def _delete_all_fields(self):
    self.labels_fields_units = []
    self.need_update = True
    
  def _new_field(self, attribute, field_class, unit, i):
    field = field_class(self.gui, self, self.o, attribute, self.undo_stack)
    self.labels_fields_units.append((attribute.label, field, unit))
    return field
    
  def get_user(self): return self.master.get_user()
  
  def get_inner_html(self):
    html  = ""
    for label, field, unit in self.labels_fields_units:
      html += """<tr><td>%s</td><td style="width: 100%%;">%s</td><td>%s</td></tr>\n""" % (html_escape(label).replace(" ", "&nbsp;"), field.get_html(), html_escape(unit))
    return html
    
  def get_html(self):
    self.need_update = False
    return """<table id="%s" border="0" style="width:100%%;">\n%s</table>\n""" % (self._id, self.get_inner_html())
    
  def get_ready_js(self):
    jss = []
    for label, field, unit in self.labels_fields_units:
      js = field.get_ready_js()
      if js: jss.append(js)
    return "\n".join(jss)
    
  def get_update_js(self):
    if self.need_update:
      return """
BYID("%s").innerHTML="%s";
%s
""" % (self._id, js_escape(self.get_inner_html()), self.get_ready_js())
    else:
      jss = []
      for label, field, unit in self.labels_fields_units:
        if field.get_need_update(): jss.append(field.get_update_js())
      return "\n".join(jss)
      
  def clear_update(self):
    self.need_update = False
    for label, field, unit in self.labels_fields_units: field.clear_update()
    
class HTMLIconPane(IconPane):
  def __init__(self, gui, master, edit_child = None, undo_stack = None):
    super().__init__(gui, master)
    self._id = master.get_user().obj2id(self)
    self.need_update = 0
    self.icon_filename = ""
    self.label         = ""
    self.details       = ""
    
  def _set_icon_filename_label_details(self, icon_filename, label, details):
    if (icon_filename != self.icon_filename) or (label != self.label) or (details != self.details):
      self.icon_filename = icon_filename
      self.label         = label
      self.details       = details
      self.need_update   = True
      
  def get_html(self):
    return u"""<span id="%s">%s</span>""" % (self._id, self.get_inner_html())
    
  def get_inner_html(self):
    icon_filename = self.descr.icon_filename_for(self.o)
    if isinstance(icon_filename, list):
      icon = """<div style="padding: 1em; float: left;">"""
      x = y = 0
      for filename in icon_filename:
        icon += """<img src="%s" style="position: absolute; left: %spx; top: %spx;"/>""" % (load_big_icon(filename), x, y)
        x += 30
        y += 15
      icon += """<img src="%s" style="margin-left: %spx; margin-top: %spx; visibility: hidden;"/>""" % (load_big_icon(filename), x - 30, y - 30)
      icon += """</div>"""
    elif icon_filename:
      icon = """<img src="%s" style="float: left; padding-right: 1em;"/>""" % load_big_icon(icon_filename)
    else: icon = ""
    return """
%s
<div style="padding-left: 10em;">
<h3>%s</h3>
<p>%s</p>
</div>
""" % (icon, html_escape(self.label), html_escape(self.details))
    
  def get_ready_js(self): return ""
  def get_need_update(self): return self.need_update

  def get_update_js(self):
    if self.need_update: return """BYID("%s").innerHTML = "%s";""" % (self._id, js_escape(self.get_inner_html()))
    return ""
    
  def clear_update(self): self.need_update = False
    
  
class HTMLChildhoodPane(ChildhoodPane):
  def __init__(self, gui, master, undo_stack = None, restrict_to_attribute = None):
    super().__init__(gui, master, undo_stack, restrict_to_attribute)
    self._id = master.get_user().obj2id(self)
    self.need_update = False
    self.visibles    = None
    
  def get_user(self): return self.master.get_user()
  
  def set_button_visibilities(self, visible1, visible2, visible3, visible4):
    if self.visibles != (visible1, visible2, visible3, visible4):
      self.visibles = visible1, visible2, visible3, visible4
    self.need_update = True
    
  def get_toolbar_items(self):
    return u"""items: [
{ type: "button", id: "up", caption: "", img: "icon-up"%s},
{ type: "button", id: "down", caption: "", img: "icon-down"%s},
{ type: "break", id: "break0"},
{ type: "button", id: "add", caption: "", img: "icon-add2"%s},
{ type: "button", id: "remove", caption: "", img: "icon-remove"%s},
], onClick : function(event) {
if (event.subItem != null)
     ws.send(JSON.stringify(["tbclick", "%s", event.subItem.text]));
else ws.send(JSON.stringify(["tbclick", "%s", event.target]));
},
""" % (", disabled: true" if not self.visibles[0] else "",
       ", disabled: true" if not self.visibles[3] else "",
       ", disabled: true" if not self.visibles[1] else "",
       ", disabled: true" if not self.visibles[2] else "",
       self._id, self._id)
    
  def get_html(self): return u"""<div id="%s" style="border: 1px solid #BBB; border-top-left-radius: 3px; border-top-right-radius: 3px;"></div>""" % (self._id)
  
  def get_ready_js(self):
    return """
tb%s = $("#%s").w2toolbar({name:"tb%s",%s});
tb%s.tree = tree%s;
""" % (self._id, self._id, self._id, self.get_toolbar_items(), self._id, self.hierarchy_pane._id)
    
  def get_need_update(self): return self.need_update
  
  def get_update_js(self):
    if self.need_update:
      js = """
tb%s.%s("up");
tb%s.%s("add");
tb%s.%s("remove");
tb%s.%s("down");
""" % (self._id, "enable" if self.visibles[0] else "disable",
       self._id, "enable" if self.visibles[1] else "disable",
       self._id, "enable" if self.visibles[2] else "disable",
       self._id, "enable" if self.visibles[3] else "disable") 
      add_actions = self.get_add_actions()
      if len(add_actions) == 1: js += """tb%s.set("add", {type:"button"});""" % self._id
      else:
        js += """
tb%s.set("add", {type:"menu", items: [%s]});
""" % (self._id, ",".join("""{text:"%s"}""" % js_escape(add_action.label) for add_action in add_actions))
        
      return js
    return ""
    
  def clear_update(self): self.need_update = False
  
  def on_click(self, action):
    if   action == "add":    self.on_add()
    elif action == "remove": self.on_remove()
    elif action == "up":     self.on_move_up()
    elif action == "down":   self.on_move_down()
    else: # Submenu add action
      add_actions = self.get_add_actions()
      for add_action in add_actions:
        if add_action.label == action:
          self.hierarchy_pane._action_activated(add_action, self.parent_o, self.attribute, self.o)
          break

class DynamicNode(object):
  def __init__(self, parent):
    self.children         = []
    self.children_created = False
    self.is_expandable    = False
    if isinstance(parent, DynamicNode):
      self.parent         = parent
    else:
      self.parent         = None
      self.update_children()
    self.update_js = ""
    self.getting_node = False
    
  def has_children(self): raise NotImplementedError
  def create_children(self, old_children = None): return []
  
  def expand(self):
    # XXX
    pass
    
  def expanded(self):
    if not self.children_created:
      self.children_created = True
      self.getting_node = True
      self.update_children()
      self.getting_node = False
    return self.children
    
  def collapsed(self): pass
      
  def update(self): pass
  
  def update_children(self):
    if self.children_created:
      old_children = self.children[:]
      new_children = self.create_children(old_children)
      old_children_set = set(old_children)
      new_children_set = set(new_children)
      
      for child in old_children[::-1]:
        if not child in new_children_set:
          if not self.getting_node: self.update_js += """tree%s.remove("%s");\n""" % (self.hierarchy_pane._id, child._id)
          child.destroy()
          
      kept_children = [child for child in new_children if child in old_children_set]
      
      for i in range(len(self.children)): self.children[i].index     = i
      for i in range(len(kept_children)): kept_children[i].new_index = i
      
      if kept_children:
        for child in kept_children:
          if child.new_index == child.index : continue # Swap with itself
          swap_with = self.children[child.new_index]
          self.update_js += """swapnode(tree%s,"%s","%s","%s");\n""" % (self.hierarchy_pane._id, self._id, child._id, swap_with._id)
          swap_with.index = child.index
          child.index     = child.new_index
          self.children[swap_with.index] = swap_with
          self.children[child    .index] = child
          
        for child in kept_children: del child.new_index
        
      for i in range(len(new_children)):
        new_children[i].index = i
        if not new_children[i] in old_children_set:
          if len(self.children) > i: before_id = '"%s"' % self.children[i]._id
          else:                      before_id = "null"
          self.children.insert(i, new_children[i])
          if not self.getting_node:
            if self.flat_list:
              if before_id == "null": self.update_js += """tree%s.add(%s);\n""" % (self.hierarchy_pane._id, new_children[i].get_json())
              else:                   self.update_js += """tree%s.insert(%s , %s);\n""" % (self.hierarchy_pane._id, before_id, new_children[i].get_json())
            else:
              self.update_js += """tree%s.insert("%s", %s, %s);\n""" % (self.hierarchy_pane._id, self._id, before_id, new_children[i].get_json())
            self.update_js += """create_icon_style(tree%s.get("%s").img);\n""" % (self.hierarchy_pane._id, new_children[i]._id)
            
      if self.update_js: self.update_js += "\n"
      
      for child in new_children:
        if not child in old_children_set:
          child.update_children()
          
      self.children = new_children
      
    else:
      if self.flat_list: pass
      else:
        is_expandable = bool(self.has_children())
        # XXX
        #if   is_expandable and (not self.is_expandable): self.tree.append(self.tree.get_iter(self.path()))
        #elif self.is_expandable and (not is_expandable): del self.tree[self.path() + (0,)]
        self.is_expandable = is_expandable
        
  def destroy(self):
    for child in self.children[::-1]: child.destroy()
    if self.parent: self.parent.children.remove(self)
    self.hierarchy_pane.nodes_needing_update.discard(self)
    
    
class HTML_HierarchyNode(HierarchyNode, DynamicNode):
  def __init__(self, hierarchy_pane, parent_node, attribute, o, restrict_to_attribute, flat_list):
    super().__init__(hierarchy_pane, parent_node, attribute, o, restrict_to_attribute, flat_list)
    self._id = self.hierarchy_pane.get_user().obj2id(self)
    
  def get_children_node_json(self):
    self.expanded() # XXX
    json =  "[%s]" % ",".join(node.get_json() for node in self.children)
    return json

  def get_json(self, with_children = False):
    icon_filename = self.descr.icon_filename_for(self.o)
    if isinstance(icon_filename, list): icon = icon[0] # Not yet supported
    if isinstance(icon_filename, str ): icon = ',img:"icon%s"' % (load_small_icon(icon_filename).rsplit("/", 1)[-1].split(".", 1)[0])
    else:                               icon = ""
    actions = self.get_actions()
    menu    = ",".join("""{id:"%s",text:"%s"}""" % (i+1, js_escape(actions[i].label)) for i in range(len(actions)))
    if with_children:
      return """{id:"%s",expanded:true,menu:[%s],nodes:[%s],text:"%s"%s}""" % (self._id, menu, ",".join(node.get_json() for node in self.children), js_escape(str(self)), icon)
    else:
      return """{id:"%s"%s,menu:[%s],text:"%s"%s}""" % (self._id, ",plus:true" if self.has_children() else "", menu, js_escape(str(self)), icon)
    
  def get_actions(self): return self.hierarchy_pane.get_actions(self.parent.o if self.parent else None, self.attribute, self.o)
  
  def do_action_by_label(self, label):
    for action in self.get_actions():
      if action.label == label:
        self.hierarchy_pane._action_activated(action, self.parent.o if self.parent else None, self.attribute, self.o)
        break
        
  def update(self):
    self.hierarchy_pane.nodes_needing_update.add(self)
    
  def get_update_js(self):
    if (self.parent is None) and self.flat_list: return self.update_js # Not shown => update only children list
    return self.update_js + """update_node(tree%s,"%s","%s",%s);""" % (self.hierarchy_pane._id, self.parent._id if self.parent else "#", self._id, self.get_json())
    
  def clear_update(self): self.update_js = ""
  
  def __str__(self):
    if self.attribute and not self.restrict_to_attribute: attr_label = self.attribute.label
    else:                                                 attr_label = ""
    if attr_label: return '''<span size="smaller" style="color:#aaaaaa;">%s</span> %s''' % (html_escape(attr_label), html_escape(self.descr.label_for(self.o)))
    return html_escape(self.descr.label_for(self.o))
    
  


EDITOR_JS += """
created_icon_styles = new Array();
function create_icon_style(name) {
  if ((name != null) && (!created_icon_styles[name])) {
    create_css_class(name, "background: url(/editobj/icons/" + name.substring(4) + ") no-repeat center; background-size: contain;");
    created_icon_styles[name] = true;
  }
}
function create_icon_styles(nodes) {
  for (i in nodes) create_icon_style(nodes[i].img);
}
function update_node(tree, parent, id, data) {
  var previous = tree.get(id);
  tree.refresh(id);
  create_icon_style(data.img);
}
function swapnode(tree, parent_id, id1, id2) {
  parent = tree.get(parent_id);
  var i = parent.nodes.indexOf(tree.get(id1));
  var j = parent.nodes.indexOf(tree.get(id2));
  var tmp = parent.nodes[i];
  parent.nodes[i] = parent.nodes[j];
  parent.nodes[j] = tmp;
  //tree.refresh();
  tree.refresh(parent_id);
}
"""

  
class HTMLHierarchyPane(HierarchyPane):
  Node = HTML_HierarchyNode
  def __init__(self, gui, master, edit_child, undo_stack = None, restrict_to_attribute = None, flat_list = 0):
    super().__init__(gui, master, edit_child, undo_stack, restrict_to_attribute, flat_list)
    self._id                  = master.get_user().obj2id(self)
    self.tree                 = None
    self.selected_parent      = None
    self.selected_attribute   = None
    self.selected_node        = None
    self.selected             = None
    self.nodes_needing_update = set()
    
  def get_user(self): return self.master.get_user()
  
  def get_html(self, div = True):
    return """<div id="%s" style="width: 100%%; height: 10em; border-left: 1px solid #BBB; border-right: 1px solid #BBB; border-bottom: 1px solid #BBB; border-bottom-left-radius: 3px; border-bottom-right-radius: 3px;">DIV</div>""" % self._id

  def get_inner_html(self): return ""
  
  READY_JS_TEMPLATE = string.Template("""
if (BYID("$id")) $tree = $$("#$id");
else             $tree = $$();
  $tree = $tree.w2sidebar({ name: "$tree", nodes: [$nodes] });
$tree.setMultiple(true);
$tree.onExpand = function(parent, data) {
  if (!$tree.get(parent).nodes.length)
    ws_send_and_wait(["getnode", "$id", parent], function (nodes) {
      create_icon_styles(nodes);
      $tree.add(parent, nodes);
    });
};
$tree.onClick = function(event) {
  event.onComplete = function() {
    var i, json = ["editnodes", "$id"];
    for(i = 0; i < $tree.selected.length; i++) { json.push($tree.selected[i]); }
    ws.send(JSON.stringify(json));
  };
};
$tree.onContextMenu = function(event) { $tree.menu = $tree.get(event.target).menu; };
$tree.onMenuClick = function(event) {
  ws.send(JSON.stringify(["nodeaction", event.target, $tree.menu[event.menuIndex].text]));
};
create_icon_styles($tree.nodes);
create_icon_styles($tree.nodes[0].nodes);
""")
  def get_ready_js(self):
    if self.flat_list: nodes = ",".join(node.get_json() for node in self.root_node.children)
    else:              nodes = self.root_node.get_json(True)
    return self.READY_JS_TEMPLATE.substitute(id = self._id, tree = "tree%s" % self._id, nodes = nodes)
    
  def get_need_update(self): return self.nodes_needing_update
  
  def get_update_js(self):
    if self.nodes_needing_update: return "\n".join(node.get_update_js() for node in self.nodes_needing_update)
    return ""
    
  def clear_update(self):
    for node in self.nodes_needing_update: node.clear_update()
    self.nodes_needing_update = set()
    
  def edit(self, o):
    if o is self.o: return
    
    HierarchyPane.edit(self, o)
    
    self.root_node.expanded()
    if self.flat_list:
      self.childhood_pane.edit(None, None, self.o)
    else:
      self.select_node(self.root_node)
      
  def expand_tree_at_level(self, level):
    pass # XXX
    
  def select_node(self, node):
    pass # XXX
    
  def on_selection_changed(self, selected_nodes):
    if len(selected_nodes) == 0:
      self.edit_child(None)
      if self.childhood_pane: self.childhood_pane.edit(None, None, None)
      self.selected_node = None
      return
      
    if len(selected_nodes) == 1:
      node = self.selected_node = selected_nodes[0]
      if node.parent: self.selected_parent = node.parent.o
      else:           self.selected_parent = None
      self.selected           = node.o
      self.selected_attribute = node.attribute
      
      self.edit_child(self.selected, self.get_ignored_attrs([node]))
      if self.childhood_pane: self.childhood_pane.edit(self.selected_parent, self.selected_attribute, self.selected)
      
    else:
      nodes = self.selected_node = selected_nodes
      self.selected_parent    = introsp.ObjectPack([(node.parent and node.parent.o) for node in nodes])
      self.selected           = introsp.ObjectPack([node.o for node in nodes])
      self.selected_attribute = [node.attribute for node in nodes]
      
      self.edit_child(self.selected, self.get_ignored_attrs(nodes))
      if self.childhood_pane: self.childhood_pane.edit(self.selected_parent, self.selected_attribute, self.selected)
  




import os, io, PIL.Image as Image

SMALL_ICON_SIZE = 32

_small_icons   = {}
_big_icons     = {}
url_2_icon     = {}
_next_image_id = 1

def guess_image_format(filename):
  filename = filename.lower()
  if filename.endswith(".svg"):  return "image/svg+xml"
  if filename.endswith(".png"):  return "image/png"
  if filename.endswith(".jpeg"): return "image/jpeg"
  if filename.endswith(".jpg"):  return "image/jpeg"
  return "image/xxx"

def load_big_icon(filename):
  url = _big_icons.get(filename)
  if not url:
    global _next_image_id
    ext = filename.rsplit(".")[-1]
    url = _big_icons[filename] = "/editobj/icons/%s" % _next_image_id
    url_2_icon[url] = guess_image_format(filename), filename
    _next_image_id += 1
  return url

def load_small_icon(filename):
  url = _small_icons.get(filename)
  if not url:
    global _next_image_id
    ext = filename.rsplit(".")[-1]
    url = _small_icons[filename] = "/editobj/icons/%s" % _next_image_id
    format = guess_image_format(filename)
    _next_image_id += 1
    if format == "image/svg+xml": # No resize needed
      url_2_icon[url] = format, filename
    else:
      image = Image.open(filename)
      w, h = image.size
      if (w > SMALL_ICON_SIZE) or (h > SMALL_ICON_SIZE):
        if w > h: image.thumbnail((SMALL_ICON_SIZE, int(float(SMALL_ICON_SIZE) * h / w)), Image.ANTIALIAS)
        else:     image.thumbnail((int(float(SMALL_ICON_SIZE) * w / h), SMALL_ICON_SIZE), Image.ANTIALIAS)
      buf = io.BytesIO()
      image.save(buf, "JPEG" if format == "image/jpeg" else "PNG")
      url_2_icon[url] = format, buf.getvalue()
  return url

load_small_icon(os.path.join(editobj3._ICON_DIR, "add.svg")) # ID 1
load_small_icon(os.path.join(editobj3._ICON_DIR, "remove.svg")) # ID 2
load_small_icon(os.path.join(editobj3._ICON_DIR, "up.svg")) # ID 3
load_small_icon(os.path.join(editobj3._ICON_DIR, "down.svg")) # ID 4



# -*- coding: utf-8 -*-
# Ontopy
# Copyright (C) 2007-2014 Jean-Baptiste LAMY
# LIMICS (Laboratoire d'informatique médicale et d'ingénierie des connaissances en santé), UMR_S 1142
# University Paris 13, Sorbonne paris-Cité, Bobigny, France

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import editobj3, editobj3.introsp as introsp, editobj3.observe as observe, editobj3.field as field

class Sum(object):
  def __init__(self, parent = None):
    self.a = 0.0
    self.b = 0.0
  def get_result(self): return self.a + self.b

class Substract(object):
  def __init__(self, parent = None):
    self.a = 0.0
    self.b = 0.0
  def get_result(self): return self.a - self.b
  
class Multiply(object):
  def __init__(self, parent = None):
    self.a = 0.0
    self.b = 0.0
  def get_result(self): return self.a * self.b
  
class Divide(object):
  def __init__(self, parent = None):
    self.a = 0.0
    self.b = 1.0
  def get_result(self): return self.a / self.b


class Calculator(object):
  def __init__(self):
    self.operations = [
      Sum(),
      Substract(),
      Multiply(),
      Divide(),
      ]

descr = introsp.description(Sum)
descr.set_label("Sum")
descr.def_attr("c", field.EnumListField([0.0, 1.0, 2.0, 3.0]))
descr.def_attr("obj", field.ObjectSelectorField, addable_values = lambda o: calculator.operations)

descr = introsp.description(Substract)
descr.set_label("Substract")

descr = introsp.description(Multiply)
descr.set_label("Multiply")

descr = introsp.description(Divide)
descr.set_label("Divide")


descr = introsp.description(Calculator)
descr.set_label("Calculator")
descr.set_details("A calculator with Ontopy.Editobj!")
#descr.def_attr("operations", field.HierarchyAndObjectListField, addable_values = [introsp.Constructor(Sum), introsp.Constructor(Substract), introsp.Constructor(Multiply), introsp.Constructor(Divide)])
#descr.def_attr("operations", field.HierarchyAndObjectListField, addable_values = [introsp.Constructor(Sum, lambda c: Sum()), introsp.Constructor(Substract, lambda c: Substract()), introsp.Constructor(Multiply, lambda c: Multiply()), introsp.Constructor(Divide, lambda c: Divide())])
descr.def_attr("operations", field.HierarchyAndObjectListField, addable_values = [introsp.NewInstanceOf(Sum), introsp.NewInstanceOf(Substract), introsp.NewInstanceOf(Multiply), introsp.NewInstanceOf(Divide)], label = "")


calculator = Calculator()

editobj3.GUI = "HTML"

import editobj3.html_utils
import editobj3.html_server
import editobj3.editor

jiba = editobj3.html_utils.User("jiba", "test")

editor = editobj3.editor.EditorDialog("HTML", jiba)
editor.edit(calculator)

jiba.register_url("/editobj/calculator.html", editor)

editor.main(path = "/editobj/calculator.html")

#editor = ontopy.editobj.editor.EditorTabbedDialog("HTML", ontopy.editobj.html_utils.GUEST)
#editor.add_tab(1, "Tab n°1", calculator)
#editor.add_tab(2, "Tab n°2", calculator.operations[0])

#ontopy.editobj.html_server.EditobjHTTPServer(("localhost", 8080)).serve_forever()

#ontopy.editobj.html_server.EditobjServer(("127.0.0.1", 8080)).serve_forever()



# -*- coding: utf-8 -*-
# Editobj3
# Copyright (C) 2007-2014 Jean-Baptiste LAMY

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import editobj3, editobj3.introsp as introsp, editobj3.observe as observe, editobj3.field as field

class Sum(object):
  def __init__(self):
    self.a = 0.0
    self.b = 0.0
  def get_result(self): return self.a + self.b

class Substract(object):
  def __init__(self):
    self.a = 0.0
    self.b = 0.0
  def get_result(self): return self.a - self.b
  
class Multiply(object):
  def __init__(self):
    self.a = 0.0
    self.b = 0.0
  def get_result(self): return self.a * self.b
  
class Divide(object):
  def __init__(self):
    self.a = 0.0
    self.b = 1.0
  def get_result(self): return self.a / self.b


class Calculator(object):
  def __init__(self):
    self.operations = [
      Sum(),
      Substract(),
      Multiply(),
      Divide(),
      ]

descr = introsp.description(Sum)
descr.set_label("Sum")

descr = introsp.description(Substract)
descr.set_label("Substract")

descr = introsp.description(Multiply)
descr.set_label("Multiply")

descr = introsp.description(Divide)
descr.set_label("Divide")


descr = introsp.description(Calculator)
descr.set_label("Calculator")
descr.set_details("A calculator with Editobj3!")
descr.def_attr("operations", label = "")


calculator = Calculator()


if len(sys.argv) > 1: editobj3.GUI = sys.argv[-1]

editobj3.edit(calculator, on_close = sys.exit).main()




# -*- coding: utf-8 -*-
# Editobj3
# Copyright (C) 2007-2014 Jean-Baptiste LAMY

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import editobj3
from editobj3.field      import *
from editobj3.html_utils import *
from editobj3.field import _WithButtonField, _RangeField, _ShortEnumField, _LongEnumField, _EnumListField

FIELD_JS = ""

class HTMLField(MultiGUIField):
  y_flags = "fill"
  def __init__(self, gui, master, o, attribute, undo_stack, *args):
    super().__init__(gui, master, o, attribute, undo_stack, *args)
    self._id = master.get_user().obj2id(self)
    self.need_update = 0
    
  def get_user(self): return self.master.get_user()
  
  def get_html       (self): return ""
  def get_ready_js   (self): return ""
  def get_update_js  (self): return ""
  def clear_update   (self): self.need_update = 0
  def get_need_update(self): return self.need_update
  def update         (self):
    if not hasattr(self, "_id"): return # we are updating during __init__
    self.need_update = 1
    

class HTMLHiddenField(HTMLField, HiddenField): pass
  
class HTMLLabelField(HTMLField, LabelField):
  def __init__(self, gui, master, o, attribute, undo_stack):
    super().__init__(gui, master, o, attribute, undo_stack)
    
  def get_html(self):
    return """<span id="%s">%s</span>""" % (self._id, html_escape(str(self.get_value())))
  
  def get_update_js(self):
    return """BYID("%s").html = "%s";""" % (self._id, js_escape(self.get_value()))
    

FIELD_JS += """function on_Field_changed(self) { ws.send(JSON.stringify(["setattr", self.id, self.value])); };\n"""

class HTMLEntryField(HTMLField, EntryField):
  def get_html(self): 
    return """<input id="%s" type="text" value="%s" onchange="on_Field_changed(this);" style="width:100%%;"/>""" % (self._id, html_escape_arg(self.get_value()))
    
  def get_update_js(self):
    return """BYID("%s").value = "%s";""" % (self._id, js_escape(self.get_value()))
    
class HTMLTextField(HTMLField, TextField):
  y_flags = "expand"
  def get_html(self):
    return """<textarea id="%s" onchange="on_Field_changed(this);" style="width:100%%;">%s</textarea>""" % (self._id, html_escape(str(self.get_value())))
    
  def get_update_js(self):
    return """BYID("%s").value = "%s";""" % (self._id, js_escape(self.get_value()))
    
    
class HTMLIntField   (HTMLEntryField, IntField   ): pass # XXX no "spin-button" since they don't allow entering e.g. "1 + 2" as an integer !
class HTMLFloatField (HTMLEntryField, FloatField ): pass
class HTMLStringField(HTMLEntryField, StringField): pass

class HTMLPasswordField(HTMLStringField, PasswordField):
  def get_html(self): 
    return """<input id="%s" type="password" value="%s" onchange="on_Field_changed(this);" style="width:100%%;"/>""" % (self._id, html_escape_arg(self.get_value()))


class HTMLEntryListField (HTMLTextField, EntryListField ): pass
class HTMLIntListField   (HTMLTextField, IntListField   ): pass
class HTMLFloatListField (HTMLTextField, FloatListField ): pass
class HTMLStringListField(HTMLTextField, StringListField): pass

def with_button_html(widget, on_button, button_text):
  return u"""<table cellspacing="0" style="width: 100%%;"><tr style="width: 100%%;"><td style="padding: 0px; width: 78%%">%s<td onclick="%s" class="editobj-button" style="margin-left: 20px; width: 20%%;"/>%s</td></tr></table>""" % (widget, on_button, button_text)

FIELD_JS += """function on_Button_clicked(id) { ws.send(JSON.stringify(["click", id])); };\n"""

class HTML_WithButtonField(HTMLField, _WithButtonField):
  #def __init__(self, gui, master, o, attribute, undo_stack, field_class = None, button_text = None, on_button = None):
  #  super().__init__(gui, master, o, attribute, undo_stack, field_class, button_text, on_button)
  def get_html(self): 
    return with_button_html(self.string_field.get_html(), """on_Button_clicked("%s")""" % self._id, html_escape_arg(editobj3.TRANSLATOR(self.button_text)))
    #return """%s<input id="%s" type="button" value="%s" onclick="on_Button_clicked(this.id);"/>""" % (self.string_field.get_html(), self._id, html_escape_arg(editobj3.TRANSLATOR(self.button_text)))
    
  def update(self): self.string_field.update()
  
  def get_need_update(self): return self.string_field.get_need_update() # Needed because attribute pane need_update can be false while some of its fields need update
  
  def get_update_js(self): return self.string_field.get_update_js()
    
  def clear_update(self):
    super().clear_update()
    self.string_field.clear_update()
    
  def on_click(self): self.on_button(self.o, self, self.undo_stack)
  
class HTMLFilenameField(HTML_WithButtonField, FilenameField):
  def on_button(self, o, field, undo_stack):
    pass # XXX
    
class HTMLDirnameField(HTML_WithButtonField, DirnameField):
  def on_button(self, o, field, undo_stack):
    pass # XXX
    
class HTMLURLField(HTML_WithButtonField, URLField):
  def on_button(self, o, field, undo_stack):
    self.get_user().add_update("""window.open("%s", "_blank");""" % (js_escape(self.get_value())))
    

FIELD_JS += """function on_Bool_changed(self) { ws.send(JSON.stringify(["setattr", self.id, self.checked])); };\n"""

class HTMLBoolField(HTMLField, BoolField):
  def get_value(self):
    if super().get_value(): return "checked"
    else:                 return ""
    
  def get_html(self):
    if self.get_value() == "checked": v = ' checked="checked"'
    else:                             v = ""
    return """<input id="%s" type="checkbox"%s onchange="on_Bool_changed(this);"/>""" % (self._id, v)
    
  def get_update_js(self):
    return """BYID("%s").checked = "%s";""" % (self._id, self.get_value())
    

class HTMLProgressBarField(HTMLField, ProgressBarField):
  def get_value(self): return "%s%%" % int(round(super().get_value() * 100.0))
  
  def get_html(self):
    return """<span id="%s">%s</span>""" % (self._id, html_escape(str(self.get_value())))
    
  def get_update_js(self):
    return """BYID("%s").html = "%s";""" % (self._id, self.get_value())
    
    
class HTML_RangeField(HTMLField, _RangeField):
  def __init__(self, gui, master, o, attribute, undo_stack, min, max, incr = 1):
    super().__init__(gui, master, o, attribute, undo_stack, min, max, incr)
    self.min = min
    self.max = max
    
  def get_html(self): 
    return u"""<input id="%s" type="range" value="%s" min="%s" max="%s" onchange="on_Field_changed(this);" style="width:100%%;"/>""" % (self._id, self.get_value(), self.min, self.max)
    
  def get_update_js(self):
    return """BYID("%s").value = "%s";""" % (self._id, self.get_value())
    
    
class HTML_ShortEnumField(HTMLField, _ShortEnumField):
  def set_value(self, value): super().set_value(self.choices[self.choice_keys[int(value)]])
    
  def get_html(self):
    index = self.choice_2_index.get(self.get_value())
    options = "".join("""<option value="%s"%s>%s</option>""" % (i, ' selected="selected"' if i == index else "", html_escape(self.choice_keys[i])) for i in range(len(self.choice_keys)))
    return u"""<select id="%s" onchange="on_Field_changed(this);" style="width:100%%; background-color: #FFFFFF;">%s</select>""" % (self._id, options)
    
  def get_update_js(self):
    return """BYID("%s").value = "%s";""" % (self._id, self.choice_2_index.get(self.get_value()))
    
class HTML_LongEnumField(HTMLField, _LongEnumField):
  y_flags = "expand"
  def set_value(self, value): super().set_value(self.choices[self.choice_keys[int(value)]])
    
  def get_html(self):
    index   = self.choice_2_index.get(self.get_value())
    options = "".join("""<option value="%s"%s>%s</option>""" % (i, ' selected="selected"' if i == index else "", html_escape(self.choice_keys[i])) for i in range(len(self.choice_keys)))
    size    = min(6, len(self.choices))
    return u"""<select id="%s" size="%s" onchange="on_Field_changed(this);" style="width:100%%; background-color: #FFFFFF;">%s</select>""" % (self._id, size, options)

  def get_update_js(self):
    return """BYID("%s").value = "%s";""" % (self._id, self.choice_2_index.get(self.get_value()))
    


FIELD_JS += """
function on_EnumList_changed(self) {
  var values = new Array();
  for (i in self.options) {
    if (self.options[i].selected) { values.push(i); }
  }
  ws.send(JSON.stringify(["setattr", self.id, values]));
};\n"""

class HTML_EnumListField(HTMLField, _EnumListField):
  y_flags = "expand"
  def set_value(self, value): super().set_value([self.choices[self.choice_keys[int(i)]] for i in value])
  
  def get_html(self):
    index   = set(self.choice_2_index.get(value) for value in self.get_value())
    options = "".join("""<option value="%s"%s>%s</option>""" % (i, ' selected="selected"' if i in index else "", html_escape(self.choice_keys[i])) for i in range(len(self.choice_keys)))
    size    = min(6, len(self.choices))
    return u"""<select id="%s" multiple="multiple" size="%s" onchange="on_EnumList_changed(this);" style="width:100%%;">%s</select>""" % (self._id, size, options)
    
  def get_update_js(self):
    values = self.get_value()
    user   = self.get_user()
    js     = """var self = BYID("%s");""" % self._id
    if values is introsp.NonConsistent: values = []
    self.indexes = { self.choice_2_index.get(value) for value in values }
    for i in range(len(self.choices)):
      js += """\nself.options[%s].selected = %s;""" % (i, str(i in self.indexes).lower())
    return js
    

class HTMLObjectAttributeField(HTMLField, ObjectAttributeField):
  def get_html(self): return self.attribute_pane.get_html()
  
  def get_ready_js(self): return self.attribute_pane.get_ready_js()
  
  def update(self): ObjectAttributeField.update(self)
  
  def get_need_update(self): return True # Needed because attribute pane need_update can be false while some of its fields need update
  
  def get_update_js(self): return self.attribute_pane.get_update_js()
    
  def clear_update(self):
    super().clear_update()
    self.attribute_pane.clear_update()

FIELD_JS += """
function on_Selector_changed(self) {
  ws.send(JSON.stringify(["setattr", self.id, self.value]));
  self.style.background = self.options[self.value].style.background;
};\n"""

class HTMLObjectSelectorField(HTMLField, ObjectSelectorField):
  def __init__(self, gui, master, o, attribute, undo_stack):
    self.current_addable_values = None
    super().__init__(gui, master, o, attribute, undo_stack)
    
  def set_value(self, value): super().set_value(self.current_addable_values[int(value)])
  
  def get_html(self):
    self.current_addable_values = self.get_addable_values()
    value = self.get_value()
    descr = introsp.description_for_object(value)
    icon_filename = descr.icon_filename_for(value)
    if isinstance(icon_filename, list): icon = icon[0] # Not yet supported
    if isinstance(icon_filename, str ): icon = '''background:url('%s') no-repeat 1px 1px; background-size: contain; ''' % (editobj3.editor_html.load_small_icon(icon_filename))
    else:                               icon = ""
    #return u"""<select id="%s" onchange="on_Selector_changed(this);" style="%spadding-left: %spx; background-color: #FFFFFF; width:78%%;">%s</select><input type="button" value="%s" onclick="on_Button_clicked('%s');" style="width:22%%; height:100%%;"/>""" % (self._id, icon, editobj3.editor_html.SMALL_ICON_SIZE + 10, self.get_inner_html(self.current_addable_values), editobj3.TRANSLATOR("Edit..."), self._id)
    #return u"""<div><select id="%s" onchange="on_Selector_changed(this);" style="%spadding-left: %spx; background-color: #FFFFFF; width:78%%;">%s</select><div class="editobj-button" onclick="on_Button_clicked('%s');" style="width:22%%; float:right; display:inline-block;">%s</div></div>""" % (self._id, icon, editobj3.editor_html.SMALL_ICON_SIZE + 10, self.get_inner_html(self.current_addable_values), self._id, editobj3.TRANSLATOR("Edit..."))
    #return """<table cellspacing="0" style="width: 100%%;"><tr style="width: 100%%;"><td style="padding: 0px; width: 78%%"><select id="%s" onchange="on_Selector_changed(this);" style="%spadding-left: %spx; background-color: #FFFFFF; width:99%%;">%s</select><td onclick="on_Button_clicked('%s');" class="editobj-button" style="margin-left: 20px; width: 20%%;"/>%s</td></tr></table>""" % (self._id, icon, editobj3.editor_html.SMALL_ICON_SIZE + 10, self.get_inner_html(self.current_addable_values), self._id, editobj3.TRANSLATOR("Edit..."))
    return with_button_html("""<select id="%s" onchange="on_Selector_changed(this);" style="%spadding-left: %spx; background-color: #FFFFFF; width:99%%;">%s</select>""" % (self._id, icon, editobj3.editor_html.SMALL_ICON_SIZE + 10, self.get_inner_html(self.current_addable_values)), "on_Button_clicked('%s');" % self._id, editobj3.TRANSLATOR("Edit..."))
      
  def get_inner_html(self, current_addable_values):
    index = current_addable_values.index(self.get_value())
    html  = ""
    for i in range(len(current_addable_values)):
      value = current_addable_values[i]
      descr = introsp.description_for_object(value)
      icon_filename = descr.icon_filename_for(value)
      if isinstance(icon_filename, list): icon = icon[0] # Not yet supported
      if isinstance(icon_filename, str ): icon = ''' style="background:url('%s') no-repeat 1px 1px white; background-size: contain; height: %spx; padding-left: %spx; padding-top: 2px"''' % (editobj3.editor_html.load_small_icon(icon_filename), editobj3.editor_html.SMALL_ICON_SIZE + 1, editobj3.editor_html.SMALL_ICON_SIZE + 10)
      else:                               icon = ""
      html += """<option value="%s"%s%s>%s</option>""" % (i, ' selected="selected"' if i == index else "", icon, html_escape(descr.label_for(value)))
    return html
    
  def get_update_js(self):
    if not hasattr(self, "_id"): return # we are updating during __init__
    current_addable_values = self.get_addable_values()
    if current_addable_values != self.current_addable_values:
      #self.current_addable_values = current_addable_values # Not here, since update should not modify the field (since if can be displayed in several browser windows)
      js = """BYID("%s").innerHTML = "%s";""" % (self._id, js_escape(self.get_inner_html(current_addable_values)))
    else: js = ""
    index = current_addable_values.index(self.get_value())
    js += """
var self = BYID("%s");
self.value = "%s";
self.style.background = self.options[self.value].style.background;
""" % (self._id, index)
    return js
    
  def clear_update(self):
    super().clear_update()
    self.current_addable_values = self.get_addable_values() # Not here, since update should not modify the field (since if can be displayed in several browser windows)
    
  def on_click(self):
    # XXX
    editobj3.edit(self.get_value(), master = self, undo_stack = self.undo_stack)
    
  def _value_listener(self, obj, type, new, old):
    self.update()
    
  def __del__(self):
    observe.unobserve(self.get_value(), self._value_listener)
    

class HTMLObjectListField(HTMLField, ObjectListField):
  y_flags = "expand"
  def __init__(self, gui, master, o, attribute, undo_stack):
    super().__init__(gui, master, o, attribute, undo_stack)
    
  def get_html(self): return """<div id="%s">%s</div>""" % (self._id, self.get_inner_html())
  
  def get_inner_html(self): return """%s%s""" % (self.childhood_pane.get_html(), self.hierarchy_pane.get_html())

  def get_ready_js(self):
    return self.hierarchy_pane.get_ready_js() + self.childhood_pane.get_ready_js()
    
  def update(self): ObjectListField.update(self)
  
  def get_need_update(self): return self.hierarchy_pane.get_need_update() or self.childhood_pane.get_need_update()
 # Needed because attribute pane need_update can be false while some of its fields need update
  
  def get_update_js(self): return self.hierarchy_pane.get_update_js() + self.childhood_pane.get_update_js()
  
  def clear_update(self):
    super().clear_update()
    self.hierarchy_pane.clear_update()
    self.childhood_pane.clear_update()



# -*- coding: utf-8 -*-
# Editobj3
# Copyright (C) 2007-2014 Jean-Baptiste LAMY

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import PyQt5.QtCore    as qtcore
import PyQt5.QtWidgets as qtwidgets

import editobj3
from editobj3.field import *
from editobj3.field import _WithButtonField, _RangeField, _ShortEnumField, _LongEnumField, _EnumListField

def _edit_from_field(o, widget, undo_stack):
  editobj3.edit(o, undo_stack = undo_stack)

class QtField(MultiGUIField):
  #y_flags = gtk.AttachOptions.FILL
  pass

class QtHiddenField(QtField, HiddenField): pass

class QtLabelField(QtField, LabelField, qtwidgets.QLabel):
  def __init__(self, gui, master, o, attribute, undo_stack):
    qtwidgets.QLabel.__init__(self)
    super().__init__(gui, master, o, attribute, undo_stack)
    self.setWordWrap(True)
    self.update()
    
  def update(self):
    self.setText(self.get_value())
    
class QtEntryField(QtField, EntryField, qtwidgets.QLineEdit):
  def __init__(self, gui, master, o, attribute, undo_stack):
    qtwidgets.QLineEdit.__init__(self)
    super().__init__(gui, master, o, attribute, undo_stack)
    
    self.update()
    self.editingFinished.connect(self.validate)
    
  def validate(self):
    self.set_value(self.text())
    
  def update(self):
    self.updating += 1
    try: self.setText(self.get_value() or "") # Qt does not accept None as a text
    finally: self.updating -= 1
    
class QtTextField(QtField, TextField, qtwidgets.QPlainTextEdit):
  #y_flags = gtk.AttachOptions.FILL | gtk.AttachOptions.EXPAND
  def __init__(self, gui, master, o, attribute, undo_stack):
    qtwidgets.QPlainTextEdit.__init__(self)
    super().__init__(gui, master, o, attribute, undo_stack)
    self.update()
    #self.textChanged.connect(self.validate)
    
  def focusOutEvent(self, event):
    qtwidgets.QPlainTextEdit.focusOutEvent(self, event)
    self.validate()
    
  def validate(self, *args):
    value = self.document().toPlainText()
    if value != self.get_value(): self.set_value(value)
    
  def update(self):
    self.updating += 1
    try:
      value = self.get_value()
      self.document().setPlainText(value)
    finally: self.updating -= 1
    
    
class QtIntField   (QtEntryField, IntField   ): pass
class QtFloatField (QtEntryField, FloatField ): pass
class QtStringField(QtEntryField, StringField): pass


class QtPasswordField(QtStringField, PasswordField):
  def __init__(self, gui, master, o, attribute, undo_stack):
    QtStringField.__init__(self, gui, master, o, attribute, undo_stack)
    self.setEchoMode(qtwidgets.QLineEdit.Password)

class QtEntryListField (QtTextField, EntryListField ): pass
class QtIntListField   (QtTextField, IntListField   ): pass
class QtFloatListField (QtTextField, FloatListField ): pass
class QtStringListField(QtTextField, StringListField): pass

class Qt_WithButtonField(QtField, _WithButtonField, qtwidgets.QWidget):
  def __init__(self, gui, master, o, attribute, undo_stack, field_class = None, button_text = None, on_button = None):
    qtwidgets.QWidget.__init__(self)
    super().__init__(gui, master, o, attribute, undo_stack, field_class, button_text, on_button)
    layout = qtwidgets.QHBoxLayout()
    layout.setContentsMargins(0, 0, 0, 0)
    layout.addWidget(self.string_field)
    button = qtwidgets.QToolButton()
    button.setText(editobj3.TRANSLATOR(self.button_text))
    button.setAutoRaise(True)
    button.clicked.connect(self.on_button_clicked)
    layout.addWidget(button)
    self.setLayout(layout)
    
  def on_button_clicked(self, *args): self.on_button(self.o, self, self.undo_stack)
    
class QtFilenameField(Qt_WithButtonField, FilenameField):
  def on_button(self, o, field, undo_stack):
    #filename = qtwidgets.QFileDialog.getSaveFileName(self, editobj3.TRANSLATOR("Choose a filename"), self.get_value())
    filename = qtwidgets.QFileDialog.getSaveFileName(self, None, self.get_value(), options = qtwidgets.QFileDialog.DontConfirmOverwrite)[0]
    if filename:
      self.string_field.set_value(filename)
      self.string_field.update()
      
class QtDirnameField(Qt_WithButtonField, DirnameField):
  def on_button(self, o, field, undo_stack):
    filename = qtwidgets.QFileDialog.getExistingDirectory(self, None, self.get_value())[0]
    if filename:
      self.string_field.set_value(filename)
      self.string_field.update()
      
class QtURLField(Qt_WithButtonField, URLField): pass


class QtBoolField(QtField, BoolField, qtwidgets.QCheckBox):
  def __init__(self, gui, master, o, attribute, undo_stack):
    qtwidgets.QCheckBox.__init__(self, "                         ")
    super().__init__(gui, master, o, attribute, undo_stack)
    self.update()
    self.stateChanged.connect(self.validate)
    
  def validate(self, state):
    if not self.updating:
      if   state == 0: self.set_value(False)
      elif state == 2: self.set_value(True)
      elif state == 1: self.setTristate(False)
      
  def update(self):
    self.updating += 1
    try:
      v = self.get_value()
      if v is introsp.NonConsistent:
        self.setTristate(True)
        self.setCheckState(1)
      else:
        self.setTristate(False)
        if v: self.setCheckState(2)
        else: self.setCheckState(0)
    finally: self.updating -= 1

class QtProgressBarField(QtField, ProgressBarField, qtwidgets.QProgressBar):
  def __init__(self, gui, master, o, attribute, undo_stack):
    qtwidgets.QProgressBar.__init__(self)
    super().__init__(gui, master, o, attribute, undo_stack)
    self.setMinimum(0)
    self.setMaximum(100)
    self.update()
    
  def update(self):
    v = self.get_value()
    if v is introsp.NonConsistent: self.pulse()
    else: self.setValue(int(round(100.0 * v)))
    

# class Qt_RangeField(QtField, _RangeField, qtwidgets.QSlider):
#   def __init__(self, gui, master, o, attribute, undo_stack, min, max, incr = 1):
#     qtwidgets.QSlider.__init__(self, qtcore.Qt.Horizontal)
#     self.setRange(min, max)
#     self.setPageStep(incr)
#     self.setTickPosition(qtwidgets.QSlider.TicksAbove)
#     super().__init__(gui, master, o, attribute, undo_stack, min, max, incr)
#     self.update()
#     self.valueChanged.connect(self.validate)
    
#   def validate(self, new_value): self.set_value(new_value)
  
#   def update(self):
#     self.updating += 1
#     value = self.get_value()
#     try:
#       value = int(value)
#       self.setValue(value)
#     except: pass
#     finally: self.updating -= 1

class Qt_RangeField(QtField, _RangeField, qtwidgets.QWidget):
  def __init__(self, gui, master, o, attribute, undo_stack, min, max, incr = 1):
    qtwidgets.QWidget.__init__(self)
    super().__init__(gui, master, o, attribute, undo_stack, min, max, incr)

    self.label = qtwidgets.QLabel()
    
    self.slider = qtwidgets.QSlider(qtcore.Qt.Horizontal)
    self.slider.setRange(min, max)
    self.slider.setPageStep(incr)
    self.slider.setTickPosition(qtwidgets.QSlider.TicksAbove)
    
    layout = qtwidgets.QHBoxLayout()
    layout.addWidget(self.label)
    layout.addWidget(self.slider)
    self.setLayout(layout)
    
    self.update()
    self.slider.valueChanged.connect(self.validate)
    
  def validate(self, new_value):
    self.label.setText(str(new_value))
    self.set_value(new_value)
  
  def update(self):
    self.updating += 1
    value = self.get_value()
    try:
      value = int(value)
      self.slider.setValue(value)
      self.label.setText(str(value))
    except: pass
    finally: self.updating -= 1

    
class Qt_ShortEnumField(QtField, _ShortEnumField, qtwidgets.QComboBox):
  def __init__(self, gui, master, o, attribute, undo_stack, choices, value_2_enum = None, enum_2_value = None):
    qtwidgets.QComboBox.__init__(self)
    super().__init__(gui, master, o, attribute, undo_stack, choices, value_2_enum, enum_2_value)
    self.setSizeAdjustPolicy(qtwidgets.QComboBox.AdjustToMinimumContentsLengthWithIcon)
    for choice in self.choice_keys: self.addItem(choice)
    self.activated.connect(self.validate)
    self.update()
    
  def validate(self, *args):
    i = self.currentIndex()
    if i != -1: self.set_value(self.choices[self.choice_keys[i]])
    
  def update(self):
    self.updating += 1
    try:
      i = self.choice_2_index.get(self.get_value())
      if i is None: self.setCurrentIndex(-1)
      else:         self.setCurrentIndex(i)
    finally: self.updating -= 1
    
class Qt_LongEnumField(QtField, _LongEnumField, qtwidgets.QListWidget):
  #y_flags = qt.AttachOptions.FILL | qt.AttachOptions.EXPAND
  def __init__(self, gui, master, o, attribute, undo_stack, choices, value_2_enum = None, enum_2_value = None):
    qtwidgets.QListWidget.__init__(self)
    super().__init__(gui, master, o, attribute, undo_stack, choices, value_2_enum, enum_2_value)
    for choice in self.choice_keys: self.addItem(choice)
    self.update()
    self.currentItemChanged.connect(self.validate)
    
  def edit(self, *args): # Hack, since both Editobj and Qt define an "edit()" method
    if len(args) == 1: _LongEnumField.edit(self, *args)
    else:              return qtwidgets.QListWidget.edit(self, *args)
    
  def validate(self, current, previous):
    if current:
      i = self.row(current)
      enum = self.choices[self.choice_keys[i]]
      self.set_value(enum)
      
  def update(self):
    self.updating += 1
    try:
      i = self.choice_2_index.get(self.get_value())
      if not i is None: self.setCurrentItem(self.item(i))
    finally: self.updating -= 1

class Qt_EnumListField(QtField, _EnumListField, qtwidgets.QListWidget):
  #y_flags = qt.AttachOptions.FILL | qt.AttachOptions.EXPAND
  def __init__(self, gui, master, o, attribute, undo_stack, choices, value_2_enum = None, enum_2_value = None):
    qtwidgets.QListWidget.__init__(self)
    super().__init__(gui, master, o, attribute, undo_stack, choices, value_2_enum, enum_2_value)
    for choice in self.choice_keys:
      item = qtwidgets.QListWidgetItem(choice, self)
      item.setFlags(item.flags() | qtcore.Qt.ItemIsUserCheckable)
      item.setCheckState(qtcore.Qt.Unchecked)
      
    self.update()
    self.itemChanged.connect(self.on_item_toggled)
    
  def edit(self, *args): # Hack, since both Editobj and Qt define an "edit()" method
    if len(args) == 1: _LongEnumField.edit(self, *args)
    else:              return qtwidgets.QListWidget.edit(self, *args)
    
  def on_item_toggled(self, item):
    if self.updating: return
    values = self.get_value()
    value  = self.choices[self.choice_keys[self.row(item)]]
    if values is introsp.NonConsistent: return # XXX does not support yet edition of multiple object
    if   item.checkState() == qtcore.Qt.Unchecked:
      values.remove(value)
    elif item.checkState() == qtcore.Qt.Checked:
      if isinstance(values, set): values.add   (value)
      else:                       values.append(value)
      
  def update(self):
    self.updating += 1
    try:
      values = self.get_value()
      if values is introsp.NonConsistent:
        for i in range(len(self.choices)): self.item(i).setCheckState(qtcore.Qt.Unchecked)
      else:
        if values: self.indexes = { self.choice_2_index.get(value) for value in values }
        else:      self.indexes = set()
        for i in range(len(self.choices)):
          if i in self.indexes: self.item(i).setCheckState(qtcore.Qt.Checked)
          else:                 self.item(i).setCheckState(qtcore.Qt.Unchecked)
    finally: self.updating -= 1


class QtObjectAttributeField(QtField, ObjectAttributeField, qtwidgets.QFrame):
  def __init__(self, gui, master, o, attribute, undo_stack):
    super().__init__(gui, master, o, attribute, undo_stack)
    qtwidgets.QFrame.__init__(self)
    self.setFrameShadow(qtwidgets.QFrame.Sunken)
    self.setFrameShape(qtwidgets.QFrame.StyledPanel)
    layout = qtwidgets.QHBoxLayout()
    layout.setContentsMargins(3, 3, 3, 3)
    layout.addWidget(self.attribute_pane)
    self.setLayout(layout)
    
class QtObjectSelectorField(QtField, ObjectSelectorField, qtwidgets.QWidget):
  def __init__(self, gui, master, o, attribute, undo_stack):
    qtwidgets.QWidget.__init__(self)
    self.current_addable_values = None
    self.combo  = qtwidgets.QComboBox()
    self.combo.setIconSize(qtcore.QSize(editobj3.editor_qt.SMALL_ICON_SIZE, editobj3.editor_qt.SMALL_ICON_SIZE))
    self.combo.setSizeAdjustPolicy(qtwidgets.QComboBox.AdjustToMinimumContentsLengthWithIcon)
    self.button = qtwidgets.QToolButton()
    self.button.setText(editobj3.TRANSLATOR("Edit..."))
    self.button.setAutoRaise(True)
    self.button.clicked.connect(self.on_click)
    layout = qtwidgets.QHBoxLayout()
    layout.setContentsMargins(0, 0, 0, 0)
    layout.addWidget(self.combo, 1)
    layout.addWidget(self.button, 0)
    self.setLayout(layout)
    super().__init__(gui, master, o, attribute, undo_stack)
    self.combo.activated.connect(self.on_change_object)
    self.destroyed.connect(self.on_destroyed)
    
  def on_destroyed(self):
    super().destroy()
    observe.unobserve(self.get_value(), self._value_listener)
    
  def on_click(self, *args): _edit_from_field(self.get_value(), self, self.undo_stack)
  
  def on_change_object(self, *args):
    self.set_value(self.current_addable_values[self.combo.currentIndex()])
    
  def _item_for(self, obj):
    descr = introsp.description_for_object(obj)
    icon_filename = descr.icon_filename_for(obj)
    if isinstance(icon_filename, str): icon = editobj3.editor_qt.load_small_icon(icon_filename)
    else:                              icon = None
    return icon, descr.label_for(obj)
  
  def store_item_for(self, obj):
    descr = introsp.description_for_object(obj)
    icon_filename = descr.icon_filename_for(obj)
    if isinstance(icon_filename, str): icon = editobj3.editor_qt.load_small_icon(icon_filename)
    else:                              icon = None
    return descr.label_for(obj), icon
  
  def _value_listener(self, obj, type, new, old):
    text, icon = self.store_item_for(obj)
    self.combo.setItemText(self.combo.currentIndex(), text)
    self.combo.setItemIcon(self.combo.currentIndex(), icon)
    
  def update(self):
    self.updating += 1
    try:
      current_addable_values = self.get_addable_values()
      if current_addable_values != self.current_addable_values:
        self.current_addable_values = current_addable_values
        self.combo.clear()
        for o in self.current_addable_values: self.combo.addItem(*self._item_for(o))
      value = self.get_value()
      try:               index = self.current_addable_values.index(value)
      except ValueError: index = -1
      self.combo.setCurrentIndex(index)
    finally: self.updating -= 1
    

class QtObjectListField(QtField, ObjectListField, qtwidgets.QWidget):
  #y_flags = qt.AttachOptions.FILL | qt.AttachOptions.EXPAND
  def __init__(self, gui, master, o, attribute, undo_stack):
    qtwidgets.QWidget.__init__(self)
    super().__init__(gui, master, o, attribute, undo_stack)
    
    layout = editobj3.editor_qt.OverlayLayout()
    layout.addWidget(self.hierarchy_pane)
    layout.addWidget(self.childhood_pane)
    self.setLayout(layout)
    
    # Timer is needed because, when clicking on a childhood_pane's button, the hierarchy_pane is focused out before the click occurs!
    self.timer = qtcore.QTimer()
    self.timer.timeout.connect(self.hide_selection)
    self.timer.setSingleShot(False)
    self.timer.setInterval(50)
    
    # Hack
    self.hierarchy_pane.focusInEvent  = self.hierarchy_focusInEvent
    self.hierarchy_pane.focusOutEvent = self.hierarchy_focusOutEvent
    
  # For updating childhood position
  def showEvent(self, event):
    qtwidgets.QWidget.showEvent(self, event)
    self.layout().invalidate()
    
  def hierarchy_focusInEvent(self, event):
    qtwidgets.QTreeView.focusInEvent(self.hierarchy_pane, event)
    self.timer.stop()
    
  def hierarchy_focusOutEvent(self, event):
    qtwidgets.QTreeView.focusOutEvent(self.hierarchy_pane, event)
    self.timer.start()
    
  def hide_selection(self):
    focus = qtwidgets.QApplication.focusWidget()
    if isinstance(focus, qtwidgets.QScrollArea) or (focus is self.childhood_pane.button_move_up) or (focus is self.childhood_pane.button_move_down) or (focus is self.childhood_pane.button_add) or (focus is self.childhood_pane.button_remove): return
    self.hierarchy_pane.select_node(None)
    self.timer.stop()
    



# -*- coding: utf-8 -*-
# Editobj3
# Copyright (C) 2007-2014 Jean-Baptiste LAMY

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import unittest

import editobj3
import editobj3.introsp as introsp
import editobj3.observe as observe
import editobj3.field   as field


class IntrospTest(object):
  def __init__(self):
    self.a = 0
    self.x = 0
    self.y = 0
    self.z = 0

  def get_x(self): return self.x
  def set_x(self, x): self.x = x

  def getY(self): return self.y
  def setY(self, y): self.y = y

  def set_z(self, z): self.z = z

  def set_b(self, b): pass
  b = property(None, set_b)
  
  def get_c(self): return 1

class TestIntrosp(unittest.TestCase):
  def setUp(self):
    pass
    
  def test_1(self):
    d = introsp.description(IntrospTest)
    assert set(d.attributes.keys()) == set(["b", "c", "x", "y", "z"])
    assert d.attributes["b"].get_method is None
    assert d.attributes["b"].set_method is not None
    assert d.attributes["c"].get_method == "get_c"
    assert d.attributes["c"].set_method is not None
    assert d.attributes["x"].get_method == "get_x"
    assert d.attributes["x"].set_method == "set_x"
    assert d.attributes["y"].get_method == "getY"
    assert d.attributes["y"].set_method == "setY"
    assert d.attributes["z"].get_method is not None
    assert d.attributes["z"].set_method == "set_z"
    
    d.def_attr("a", field.HiddenField)
    i = IntrospTest()
    attributes = d.attributes_of(i)
    attributes = { attribute.name : attribute for attribute in attributes }
    print(attributes)
    assert attributes["a"].field_class_for(i) is field.HiddenField
    
  def test_getset_2(self):
    d = introsp.description(IntrospTest)
    i = IntrospTest()
    attributes = d.attributes_of(i)
    assert set(attribute.name for attribute in attributes) == set(['a', 'c', 'b', 'y', 'x', 'z'])
    
    

class C(object):
  pass

class TestObserve(unittest.TestCase):
  def setUp(self):
    self.listened = None
    
  def listener(self, *args): self.listened = args
  
  def test_object_1(self):
    c = C()
    observe.observe(c, self.listener)
    observe.scan()
    assert not self.listened
    c.x = 1
    observe.scan()
    assert self.listened == (c, object, {"x" : 1}, {})
  
  def test_object_2(self):
    c = C()
    observe.observe(c, self.listener)
    observe.scan()
    assert not self.listened
    c.x = 1
    c.y = "e"
    observe.scan()
    assert self.listened == (c, object, {"x" : 1, "y" : "e"}, {})
    
    
  def test_list_1(self):
    c = [0]
    observe.observe(c, self.listener)
    observe.scan()
    assert not self.listened
    c.append(1)
    observe.scan()
    assert self.listened == (c, list, [0, 1], [0])
    c.remove(0)
    observe.scan()
    assert self.listened == (c, list, [1], [0, 1])
    
    
  def test_set_1(self):
    c = set([0])
    observe.observe(c, self.listener)
    observe.scan()
    assert not self.listened
    c.add(1)
    observe.scan()
    assert self.listened == (c, set, set([0, 1]), set([0]))
    c.remove(0)
    observe.scan()
    assert self.listened == (c, set, set([1]), set([0, 1]))
    
   
  def test_tree_1(self):
    c  = C()
    c2 = C()
    c.l = [c2]
    observe.observe_tree(c, self.listener)
    assert observe.isobserved(c)
    assert observe.isobserved(c2)
    observe.unobserve_tree(c, self.listener)
    assert not observe.isobserved(c)
    assert not observe.isobserved(c2)
    
  def test_tree_2(self):
    c  = C()
    c2 = C()
    c.l = (1, 2, c2)
    observe.observe_tree(c, self.listener)
    assert observe.isobserved(c)
    assert observe.isobserved(c2)
    c.l = 1
    observe.scan()
    assert observe.isobserved(c)
    assert not observe.isobserved(c2)
    
    observe.unobserve_tree(c, self.listener)
    assert not observe.isobserved(c)
    assert not observe.isobserved(c2)
    
  def test_tree_3(self):
    c  = C()
    c2 = C()
    c.l = set([1, 2, c2])
    observe.observe_tree(c, self.listener)
    assert observe.isobserved(c)
    assert observe.isobserved(c2)
    
    observe.unobserve_tree(c, self.listener)
    assert not observe.isobserved(c)
    assert not observe.isobserved(c2)
    
  def test_tree_4(self):
    c  = C()
    c2 = C()
    c.l = []
    observe.observe_tree(c, self.listener)
    c.l.append(c2)
    observe.scan()
    assert observe.isobserved(c)
    assert observe.isobserved(c2)
    c.l.remove(c2)
    observe.scan()
    assert observe.isobserved(c)
    assert not observe.isobserved(c2)
    
    observe.unobserve_tree(c, self.listener)
    assert not observe.isobserved(c)
    assert not observe.isobserved(c2)
    
  def test_tree_5(self):
    c  = C()
    c2 = C()
    c .c = c2
    c2.c = c
    observe.observe_tree(c, self.listener)
    assert observe.isobserved(c)
    assert observe.isobserved(c2)
    
    observe.unobserve_tree(c, self.listener)
    assert not observe.isobserved(c)
    assert not observe.isobserved(c2)


if __name__ == '__main__':
  unittest.main()
  



# -*- coding: utf-8 -*-
# Editobj3
# Copyright (C) 2007-2014 Jean-Baptiste LAMY

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, os, os.path, editobj3, editobj3.introsp as introsp, editobj3.observe as observe, editobj3.field as field, editobj3.undoredo as undoredo, editobj3.editor as editor

class Module(object):
  def __init__(self, module, image = None):
    self.module          = module
    self.details         = module.__doc__ or ""
    self.submodules      = []
    if image:
      self.icon_filename = image
      
  def __str__(self): return self.module.__name__

mod = Module(editobj3, os.path.join(os.path.dirname(sys.argv[0]), "./dialog.png"))
mod.url = "http://home.gna.org/oomadness/en/editobj"
mod.version = editobj3.VERSION
mod.submodules.append(Module(introsp))
mod.submodules.append(Module(observe))
mod.submodules.append(Module(undoredo))
mod.submodules.append(Module(editor))
mod.submodules.append(Module(field))


descr = introsp.description(Module)
descr.def_attr("module"       , field.HiddenField)
descr.def_attr("details"      , field.HiddenField)
descr.def_attr("icon_filename", field.HiddenField)
descr.def_attr("submodules", label = "", remove_method = None, reorder_method = None)

# The following are not needed because Editobj is smart enough to guess them;
# they are kept only for documentation purpose.

#descr.set_details(lambda o: o.details)
#descr.set_label  (lambda o: str(o))

if len(sys.argv) > 1: editobj3.GUI = sys.argv[-1]

editobj3.edit(mod, on_close = sys.exit).main()



# -*- coding: utf-8 -*-
# Editobj3
# Copyright (C) 2007-2014 Jean-Baptiste LAMY

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Editobj3 -- An automatic dialog box generator

Editobj is able to generate a dialog box for editing almost any Python object. The dialog box is
composed of an attribute list, a luxurious good-looking but useless icon and title bar, and a tree
view (if the edited object is part of a tree-like structure). The default dialog box can be
customized and adapted for a given class of object through the editobj3.introsp module.

Editobj3 has multiple GUI support; it currently supports Qt, Gtk and HTML.

Editobj3 was inspired by Java's "Bean Editor", however the intial concept have been extended by
adding icons, tree views, undo/redo, translation support, automatic update on object changes,
and support for editing several objects as if there was only one (see editobj3.introsp.ObjectPack).


The editobj3 package contains the following modules:

 * editobj3.introsp: High level, highly customizable introspection (go there for customizing Editobj3 dialog boxes)
 * editobj3.observe: Observation framework
 * editobj3.undoredo: Multiple undo/redo support
 * editobj3.editor: The editor dialog boxes and related widgets
 * editobj3.field: The set of basic fields for attribute panes


The following global variables can be changed:

 * editobj3.GUI: the default GUI system (default is Qt if available, else Gtk, else HTML)
 * editobj3.TRANSLATOR: the translator function used for translating dialog boxes.

It can be set to a translator function (such as the ones from the gettext module).


The edit() function in this module is an easy way to quickly edit an object. More complex edition
can be done using the widget available in editobj3.editor.
"""

import os, os.path

_ICON_DIR = os.path.join(os.path.dirname(__file__), "icons")
if not os.path.exists(_ICON_DIR):
  _ICON_DIR = os.path.join("/usr", "share", "editobj3", "icons")
  if not os.path.exists(_ICON_DIR):
    _ICON_DIR = os.path.join("/usr", "local", "share", "editobj3", "icons")
    if not os.path.exists(_ICON_DIR):
      _ICON_DIR = os.path.join("/usr", "share", "python-editobj3", "icons")
      
_eval = eval
def eval(s):
  return _eval(s)

VERSION = "0.2"

GUI = "Qt"

TRANSLATOR = lambda s: s

def edit(o, on_validate = None, direction = "h", undo_stack = None, width = -1, height = -1, expand_tree_at_level = False, selected = None, edit_child_in_self = True, on_close = None, on_edit_child = None, flat_list = False, master = None, menubar = True):
  global GUI
  if not GUI:
    try:
      import PyQt5
      GUI = "Qt"
    except:
      try:
        import gi
        from gi.repository import Gtk as gtk
        GUI = "Gtk"
      except:
        GUI = "HTML"
        
  if GUI == "Qt":
    import sys, PyQt5.QtWidgets as qtwidgets
    if qtwidgets.QApplication.startingUp(): qtwidgets.app = qtwidgets.QApplication(sys.argv)
    
  import editobj3.editor
  dialog = editobj3.editor.EditorDialog(GUI, master, direction, on_validate, edit_child_in_self, undo_stack, on_close, menubar)
  if flat_list: dialog.editor_pane.hierarchy_pane.flat_list = True
  if (width != -1) or (height != -1): dialog.set_default_size(width, height)
  if on_edit_child:
    dialog.editor_pane.on_edit_child = on_edit_child
  dialog.edit(o)
  if expand_tree_at_level: dialog.editor_pane.hierarchy_pane.expand_tree_at_level(expand_tree_at_level)
  if selected:
    dialog.editor_pane.hierarchy_pane.edit_child(selected)
  dialog.show()
  return dialog




# -*- coding: utf-8 -*-
# Editobj3
# Copyright (C) 2007-2014 Jean-Baptiste LAMY

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, os, os.path, stat

import editobj3, editobj3.introsp as introsp, editobj3.observe as observe, editobj3.field as field


class Permissions(object):
  def __init__(self, can_read, can_write, can_execute):
    self.can_read    = can_read
    self.can_write   = can_write
    self.can_execute = can_execute
    
  #def set_can_read(self, can_read):
    # XXX change the permission here
    # (not done in the example)
    # (and so on for the other attributes)


class File(object):
  def __init__(self, path):
    if isinstance(path, str): path = path
    self.path              = os.path.abspath(path)
    self.filename          = os.path.basename(path)
    try:    self.size = os.path.getsize(path)
    except: pass
    try:    mode = os.stat(path).st_mode
    except: mode = None
    if mode is not None:
      self.permissions_user  = Permissions(bool(mode & stat.S_IRUSR), bool(mode & stat.S_IWUSR), bool(mode & stat.S_IXUSR))
      self.permissions_group = Permissions(bool(mode & stat.S_IRGRP), bool(mode & stat.S_IWGRP), bool(mode & stat.S_IXGRP))
      self.permissions_other = Permissions(bool(mode & stat.S_IROTH), bool(mode & stat.S_IWOTH), bool(mode & stat.S_IXOTH))
      
    self.children = None
    
  def __str__(self): return self.filename
  
  def get_icon_filename(self):
    if os.path.isdir(self.path):    return os.path.join(os.path.dirname(sys.argv[0]), "directory.png")
    elif self.path.endswith(".py"): return os.path.join(editobj3._ICON_DIR, "python.svg")
    elif self.path.endswith(".png") or self.path.endswith(".jpeg") or self.path.endswith(".jpg"): return self.path
    else:                           return os.path.join(os.path.dirname(sys.argv[0]), "file.png")
    
  def has_children(self): return os.path.isdir(self.path)
  
  
  def get_children(self):
    if self.children is None:
      if os.path.isdir(self.path):
        self.children = [File(os.path.join(self.path, filename)) for filename in os.listdir(self.path) if not filename.startswith(".")]
        self.children.sort(key = lambda file: file.path)
        #self.children = range(10)
      else: self.children = []
    return self.children


descr = introsp.description(File)
descr.set_icon_filename(lambda o: o.get_icon_filename())
descr.def_attr("icon_filename", field.HiddenField)
descr.def_attr("size", field.IntField, unit = "Ko")
descr.def_attr("children", label = "")
descr.def_attr("permissions_user" , field.ObjectAttributeField)
descr.def_attr("permissions_group", field.ObjectAttributeField)
descr.def_attr("permissions_other", field.ObjectAttributeField)



file = File(os.path.join(".."))

if len(sys.argv) > 1: editobj3.GUI = sys.argv[-1]

editobj3.edit(file, on_close = sys.exit).main()




# -*- coding: utf-8 -*-
# Editobj3
# Copyright (C) 2007-2014 Jean-Baptiste LAMY

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, os, os.path, editobj3, editobj3.introsp as introsp, editobj3.observe as observe, editobj3.field as field
import editobj3.undoredo as undoredo

class User(object):
  def __init__(self, login, icon_filename, session, language):
    self.login         = login
    self.password      = ""
    self.icon_filename = icon_filename
    self.session       = session
    self.language      = language
    
  def __str__(self): return self.login


class UserSelection(object):
  def __init__(self):
    self.users = []
    
  def __str__(self): return "User selection"
  

icon_filename = os.path.join(os.path.dirname(sys.argv[0]), "./jiba.png")

user_selection = UserSelection()
user_selection.users.append(User("Jiba"    , icon_filename, "WindowMaker", "Français"))
user_selection.users.append(User("Blam"    , icon_filename, "FluxBox"    , "Français"))
user_selection.users.append(User("Marmoute", icon_filename, "Gnome"      , "Français"))


descr = introsp.description(User)
descr.def_attr("icon_filename", field.HiddenField)
descr.def_attr("session" , field.EnumField(["WindowMaker", "FluxBox", "LXDE", "Gnome", "KDE"]))
descr.def_attr("language", field.EnumField(["Français", "Italiano", "English", "Esperanto"], long_list = 1))

# The following are not needed because Editobj is smart enough to guess them;
# they are kept only for example purpose.
#descr.def_attr("password", field.PasswordField)
#descr.set_icon_filename(lambda o: o.icon_filename)

descr = introsp.description(UserSelection)
descr.set_details("1) Choose a user\n2) Type a valid password\n3) Login!")

descr.add_action(introsp.Action("Undo", lambda undo_stack, editor, o: undoredo.stack.undo()))
descr.add_action(introsp.Action("Redo", lambda undo_stack, editor, o: undoredo.stack.redo()))


def on_validate(user):
  if isinstance(user, User):
    print("%s has loged in in with password '%s', language '%s' and session type '%s'." % (user.login, user.password, user.language, user.session))
  else:
    print("User has canceled.")
  sys.exit()
  
if len(sys.argv) > 1: editobj3.GUI = sys.argv[-1]

editobj3.edit(user_selection, on_validate = on_validate).main()
